import numpy as np
import scipy
from pytest import raises

from pydmd import DMDc


def create_system_with_B():
    snapshots = np.array([[4, 2, 1, 0.5, 0.25], [7, 0.7, 0.07, 0.007, 0.0007]])
    u = np.array([-4, -2, -1, -0.5])
    B = np.array([[1, 0]]).T
    return {"snapshots": snapshots, "u": u, "B": B}


def create_system_without_B():
    n = 5  # dimension snapshots
    m = 15  # number snapshots
    A = scipy.linalg.helmert(n, True)
    B = np.random.rand(n, n) - 0.5
    x0 = np.array([0.25] * n)
    u = np.random.rand(n, m - 1) - 0.5
    snapshots = [x0]
    for i in range(m - 1):
        snapshots.append(A.dot(snapshots[i]) + B.dot(u[:, i]))
    snapshots = np.array(snapshots).T
    return {"snapshots": snapshots, "u": u, "B": B, "A": A}


def create_system_without_B_lag(lag=1):
    n = 5  # dimension snapshots
    m = 15  # number snapshots
    A = scipy.linalg.helmert(n, True)
    B = np.random.rand(n, n) - 0.5
    snapshots = []
    for _ in range(lag):
        snapshots.append(np.array([0.25] * n))
    u = np.random.rand(n, m - lag) - 0.5
    for i in range(m - lag):
        snapshots.append(A.dot(snapshots[i]) + B.dot(u[:, i]))
    snapshots = np.array(snapshots).T
    return {"snapshots": snapshots, "u": u, "B": B, "A": A}


def test_eigs_b_known():
    system = create_system_with_B()
    dmdc = DMDc(svd_rank=-1)
    dmdc.fit(system["snapshots"], system["u"], system["B"])
    real_eigs = np.array([0.1, 1.5])
    np.testing.assert_array_almost_equal(dmdc.eigs, real_eigs)


def test_eigs_b_unknown():
    system = create_system_without_B()
    dmdc = DMDc(svd_rank=3, opt=False, svd_rank_omega=4)
    dmdc.fit(system["snapshots"], system["u"])
    assert dmdc.eigs.shape[0] == 3


def test_eigs_b_unknown_lag():
    lag = 3
    system = create_system_without_B_lag(lag=lag)
    dmdc = DMDc(svd_rank=3, opt=False, svd_rank_omega=4, lag=lag)
    dmdc.fit(system["snapshots"], system["u"])
    assert dmdc.eigs.shape[0] == 3


def test_modes_b_unknown():
    system = create_system_without_B()
    dmdc = DMDc(svd_rank=3, opt=False, svd_rank_omega=4)
    dmdc.fit(system["snapshots"], system["u"])
    assert dmdc.modes.shape[1] == 3


def test_modes_b_unknown_lag():
    lag = 3
    system = create_system_without_B_lag(lag=lag)
    dmdc = DMDc(svd_rank=3, opt=False, svd_rank_omega=4, lag=lag)
    dmdc.fit(system["snapshots"], system["u"])
    assert dmdc.modes.shape[1] == 3


def test_reconstruct_b_known():
    system = create_system_with_B()
    dmdc = DMDc(svd_rank=-1)
    dmdc.fit(system["snapshots"], system["u"], system["B"])
    np.testing.assert_array_almost_equal(
        dmdc.reconstructed_data(), system["snapshots"]
    )


def test_B_b_known():
    system = create_system_with_B()
    dmdc = DMDc(svd_rank=-1)
    dmdc.fit(system["snapshots"], system["u"], system["B"])
    np.testing.assert_array_almost_equal(dmdc.B, system["B"])


def test_reconstruct_b_unknown():
    system = create_system_without_B()
    dmdc = DMDc(svd_rank=-1, opt=True)
    dmdc.fit(system["snapshots"], system["u"])
    np.testing.assert_array_almost_equal(
        dmdc.reconstructed_data(), system["snapshots"], decimal=6
    )


def test_reconstruct_b_unknown_lag():
    lag = 3
    system = create_system_without_B_lag(lag=lag)
    dmdc = DMDc(svd_rank=-1, opt=True, lag=lag)
    dmdc.fit(system["snapshots"], system["u"])
    np.testing.assert_array_almost_equal(
        dmdc.reconstructed_data(), system["snapshots"], decimal=6
    )


def test_atilde_b_unknown():
    system = create_system_without_B()
    dmdc = DMDc(svd_rank=-1, opt=True)
    dmdc.fit(system["snapshots"], system["u"])
    expected_atilde = dmdc.basis.T.conj().dot(system["A"]).dot(dmdc.basis)
    np.testing.assert_array_almost_equal(
        dmdc.operator.as_numpy_array, expected_atilde, decimal=1
    )


def test_get_bitmask_default():
    system = create_system_with_B()
    dmd = DMDc(svd_rank=-1, opt=True)
    dmd.fit(system["snapshots"], system["u"], system["B"])
    assert np.all(dmd.modes_activation_bitmask == True)


def test_set_bitmask():
    system = create_system_with_B()
    dmd = DMDc(svd_rank=-1, opt=True)
    dmd.fit(system["snapshots"], system["u"], system["B"])

    new_bitmask = np.full(len(dmd.amplitudes), True, dtype=bool)
    new_bitmask[[0]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.modes_activation_bitmask[0] == False
    assert np.all(dmd.modes_activation_bitmask[1:] == True)


def test_not_fitted_get_bitmask_raises():
    dmd = DMDc(svd_rank=-1, opt=True)
    with raises(RuntimeError):
        print(dmd.modes_activation_bitmask)


def test_not_fitted_set_bitmask_raises():
    dmd = DMDc(svd_rank=-1, opt=True)
    with raises(RuntimeError):
        dmd.modes_activation_bitmask = np.full(3, True, dtype=bool)


def test_raise_wrong_dtype_bitmask():
    system = create_system_with_B()
    dmd = DMDc(svd_rank=-1, opt=True)
    dmd.fit(system["snapshots"], system["u"], system["B"])
    with raises(RuntimeError):
        dmd.modes_activation_bitmask = np.full(3, 0.1)


def test_fitted():
    system = create_system_with_B()
    dmd = DMDc(svd_rank=-1, opt=True)
    assert not dmd.fitted
    dmd.fit(system["snapshots"], system["u"], system["B"])
    assert dmd.fitted


def test_bitmask_amplitudes():
    system = create_system_with_B()
    dmd = DMDc(svd_rank=-1, opt=True)
    dmd.fit(system["snapshots"], system["u"], system["B"])

    old_n_amplitudes = dmd.amplitudes.shape[0]
    retained_amplitudes = np.delete(dmd.amplitudes, [0, -1])

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.amplitudes.shape[0] == old_n_amplitudes - 2
    np.testing.assert_almost_equal(dmd.amplitudes, retained_amplitudes)


def test_bitmask_eigs():
    system = create_system_with_B()
    dmd = DMDc(svd_rank=-1, opt=True)
    dmd.fit(system["snapshots"], system["u"], system["B"])

    old_n_eigs = dmd.eigs.shape[0]
    retained_eigs = np.delete(dmd.eigs, [0, -1])

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.eigs.shape[0] == old_n_eigs - 2
    np.testing.assert_almost_equal(dmd.eigs, retained_eigs)


def test_bitmask_modes():
    system = create_system_with_B()
    dmd = DMDc(svd_rank=-1, opt=True)
    dmd.fit(system["snapshots"], system["u"], system["B"])

    old_n_modes = dmd.modes.shape[1]
    retained_modes = np.delete(dmd.modes, [0, -1], axis=1)

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.modes.shape[1] == old_n_modes - 2
    np.testing.assert_almost_equal(dmd.modes, retained_modes)


def test_reconstructed_data():
    system = create_system_with_B()
    dmd = DMDc(svd_rank=-1, opt=True)
    dmd.fit(system["snapshots"], system["u"], system["B"])

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    dmd.reconstructed_data
    assert True


def test_getitem_modes():
    system = create_system_with_B()
    dmd = DMDc(svd_rank=-1)
    dmd.fit(system["snapshots"], system["u"], system["B"])
    old_n_modes = dmd.modes.shape[1]

    assert dmd[[0, -1]].modes.shape[1] == 2
    np.testing.assert_almost_equal(dmd[[0, -1]].modes, dmd.modes[:, [0, -1]])

    assert dmd.modes.shape[1] == old_n_modes

    assert dmd[1::2].modes.shape[1] == old_n_modes // 2
    np.testing.assert_almost_equal(dmd[1::2].modes, dmd.modes[:, 1::2])

    assert dmd.modes.shape[1] == old_n_modes

    assert dmd[1].modes.shape[1] == 1
    np.testing.assert_almost_equal(np.squeeze(dmd[1].modes), dmd.modes[:, 1])

    assert dmd.modes.shape[1] == old_n_modes


def test_getitem_raises():
    system = create_system_with_B()
    dmd = DMDc(svd_rank=-1)
    dmd.fit(system["snapshots"], system["u"], system["B"])

    with raises(ValueError):
        dmd[[0, 1, 1, 0, 1]]
    with raises(ValueError):
        dmd[[True, True, False, True]]
    with raises(ValueError):
        dmd[1.0]


def test_correct_amplitudes():
    system = create_system_with_B()
    dmd = DMDc(svd_rank=-1)
    dmd.fit(system["snapshots"], system["u"], system["B"])
    np.testing.assert_array_almost_equal(dmd.amplitudes, dmd._b)


def test_lag_param_b_unknown_raises():
    system = create_system_without_B_lag(lag=3)
    dmdc = DMDc(svd_rank=-1, opt=True, lag=0)
    with raises(ValueError):
        dmdc.fit(system["snapshots"], system["u"])



"""
PiDMD utilities module.

References:
- Peter J. Baddoo, Benjamin Herrmann, Beverley J. McKeon, J. Nathan Kutz, and
Steven L. Brunton. Physics-informed dynamic mode decomposition (pidmd). 2021.
arXiv:2112.04307.
"""

import numpy as np
from numpy.fft import fft, fft2, ifft
from scipy import sparse
from scipy.linalg import block_diag, rq

from .utils import compute_rank, compute_svd


def compute_unitary(X, Y, svd_rank):
    """
    Given the data matrices X and Y and the rank truncation svd_rank, solves
    for the best-fit unitary operator A that solves the relationship Y = AX.
    Returns a dictionary containing the corresponding reduced operator atilde.
    """
    Ux = compute_svd(X, svd_rank)[0]
    Yproj = Ux.conj().T.dot(Y)
    Xproj = Ux.conj().T.dot(X)
    Uyx, _, Vyx = compute_svd(Yproj.dot(Xproj.conj().T), -1)
    atilde = Uyx.dot(Vyx.conj().T)
    return {"atilde": atilde}


def compute_uppertriangular(X, Y):
    """
    Given the data matrices X and Y, solves for the best-fit
    uppertriangular matrix A that solves the relationship Y = AX.
    Returns a dictionary containing A.
    """
    R, Q = rq(X, mode="economic")
    Ut = np.triu(Y.dot(Q.conj().T))
    A = np.linalg.lstsq(R.T, Ut.T, rcond=None)[0].T
    return {"A": A}


def compute_diagonal(X, Y, svd_rank, manifold_opt, compute_A):
    """
    Given the data matrices X and Y and the rank truncation svd_rank, solves
    for the best-fit matrix A that solves the relationship Y = AX and has
    diagonal entries specified by manifold_opt. Only the eigenvalues and
    eigenvectors of A are computed if the compute_A flag is not True.
    Returns a dictionary of computed values.
    """
    # Specify the index matrix for the diagonals of A.
    nx = len(X)
    if manifold_opt is None:
        ind_mat = np.ones((nx, 2), dtype=int)
    elif isinstance(manifold_opt, int) and manifold_opt > 0:
        ind_mat = manifold_opt * np.ones((nx, 2), dtype=int)
    elif (
        isinstance(manifold_opt, tuple)
        and len(manifold_opt) == 2
        and np.all(np.array(manifold_opt) > 0)
    ):
        ind_mat = np.ones((nx, 2))
        ind_mat[:, 0] *= manifold_opt[0]
        ind_mat[:, 1] *= manifold_opt[1]
    elif (
        isinstance(manifold_opt, np.ndarray)
        and manifold_opt.shape == (nx, 2)
        and np.all(manifold_opt > 0)
    ):
        ind_mat = manifold_opt
    else:
        raise ValueError("manifold_opt is not in an allowable format.")

    # Keep track of info for building A as a sparse coordinate matrix.
    I = []
    J = []
    R = []

    # Solve min||Cx-b|| along each row.
    nxs = np.arange(nx)
    l1s = (nxs - ind_mat[:, 0] + 1).clip(min=0).astype(int)
    l2s = (nxs + ind_mat[:, 1]).clip(max=nx).astype(int)
    for j in range(nx):
        l1, l2 = l1s[j], l2s[j]
        C = X[l1:l2].T
        b = Y[j].T
        I.append(j * np.ones(l2 - l1))
        J.append(np.arange(l1, l2))
        R.append(np.linalg.lstsq(C, b, rcond=None)[0].T)

    # Build A as a sparse matrix.
    A_sparse = sparse.coo_matrix(
        (np.hstack(R), (np.hstack(I), np.hstack(J))), shape=(nx, nx)
    )
    if compute_A:
        return {"A": A_sparse.toarray()}
    r = compute_rank(X, svd_rank)
    eigenvalues, modes = sparse.linalg.eigs(A_sparse, k=r)
    return {"eigenvalues": eigenvalues, "modes": modes}


def compute_symmetric(X, Y, svd_rank, skew_symmetric=False):
    """
    Given the data matrices X and Y and the rank truncation svd_rank, solves
    for the best-fit symmetric (or skew-symmetric) operator A that solves the
    relationship Y = AX. Returns a dictionary containing the corresponding
    reduced operator atilde.
    """
    U, s, V = compute_svd(X, -1)
    C = np.linalg.multi_dot([U.conj().T, Y, V])
    r = compute_rank(X, svd_rank)
    if skew_symmetric:
        atilde = 1j * np.diag(np.diagonal(C).imag / s)[:r, :r]
        for i in range(r):
            for j in range(i + 1, r):
                atilde[i, j] = -s[i] * np.conj(C[j, i]) + s[j] * C[i, j]
                atilde[i, j] /= s[i] ** 2 + s[j] ** 2
        atilde += -atilde.conj().T - 1j * np.diag(np.diag(atilde.imag))
    else:  # symmetric
        atilde = np.diag(np.diagonal(C).real / s)[:r, :r]
        for i in range(r):
            for j in range(i + 1, r):
                atilde[i, j] = s[i] * np.conj(C[j, i]) + s[j] * C[i, j]
                atilde[i, j] /= s[i] ** 2 + s[j] ** 2
        atilde += atilde.conj().T - np.diag(np.diag(atilde.real))
    return {"atilde": atilde}


def compute_toeplitz(X, Y, flipped=False):
    """
    Given the data matrices X and Y, solves for the best-fit toeplitz operator
    A (or hankel operator if flipped = True) that solves the relationship
    Y = AX. Returns a dictionary containing A.
    """
    nx, nt = X.shape

    if flipped:  # hankel
        J = np.fliplr(np.eye(nx))
    else:  # toeplitz
        J = np.eye(nx)

    # Define left and right matrices.
    Am = fft(np.hstack([np.eye(nx), np.zeros((nx, nx))]).T, axis=0)
    Am = Am.conj().T / np.sqrt(2 * nx)
    B = fft(np.hstack([J.dot(X).conj().T, np.zeros((nt, nx))]).T, axis=0)
    B = B.conj().T / np.sqrt(2 * nx)

    # Compute AA* and B*B (fast computation of AA*).
    AAt = ifft(
        fft(
            np.vstack(
                [
                    np.hstack([np.eye(nx), np.zeros((nx, nx))]),
                    np.zeros((nx, 2 * nx)),
                ]
            ),
            axis=0,
        ).T,
        axis=0,
    ).T

    # Solve linear system y = dL.
    y = np.diag(np.linalg.multi_dot([Am.conj().T, Y.conj(), B])).conj().T
    L = np.multiply(AAt, B.conj().T.dot(B).T).conj().T
    eigenvalues = np.linalg.lstsq(L[:-1, :-1].T, y[:-1], rcond=None)[0]
    eigenvalues = np.append(eigenvalues, 0)

    # Convert eigenvalues into circulant matrix.
    new_A = ifft(fft(np.diag(eigenvalues), axis=0).T, axis=0).T

    # Extract toeplitz matrix from the circulant matrix.
    A = new_A[:nx, :nx].dot(J)

    return {"A": A}


def compute_circulant(X, Y, circulant_opt, svd_rank, compute_A):
    """
    Given the data matrices X and Y and the rank truncation svd_rank, solves
    for the best-fit circulant matrix A that solves the relationship Y = AX and
    satisfies any additional conditions set by circulant_opt. Only the
    eigenvalues and eigenvectors of A are computed if the compute_A
    flag is not True. Returns a dictionary of computed values.
    """
    nx = len(X)
    fX = fft(X, axis=0)
    fY = fft(Y.conj(), axis=0)
    fX_norm = np.linalg.norm(fX, axis=1)
    eigenvalues = np.divide(np.diag(fX.dot(fY.conj().T)), fX_norm**2)

    if circulant_opt == "unitary":
        eigenvalues = np.exp(1j * np.angle(eigenvalues))
    elif circulant_opt == "symmetric":
        eigenvalues = eigenvalues.real
    elif circulant_opt == "skewsymmetric":
        eigenvalues = 1j * eigenvalues.imag

    # Remove the least important eigenvalues.
    r = compute_rank(X, svd_rank)
    res = np.divide(
        np.diag(np.abs(fX.dot(fY.conj().T))),
        np.linalg.norm(fX.conj().T, 2, axis=0).T,
    )
    ind_exclude = np.argpartition(res, nx - r)[: nx - r]
    eigenvalues[ind_exclude] = 0

    if compute_A:
        A = fft(np.multiply(eigenvalues, ifft(np.eye(nx), axis=0).T).T, axis=0)
        return {"A": A}

    eigenvalues = np.delete(eigenvalues, ind_exclude)
    modes = np.delete(fft(np.eye(nx), axis=0), ind_exclude, axis=1)
    return {"eigenvalues": eigenvalues, "modes": modes}


def compute_symtridiagonal(X, Y, svd_rank, compute_A):
    """
    Given the data matrices X and Y and the rank truncation svd_rank, solves
    for the best-fit symmetric tridiagonal matrix A that solves Y = AX.
    Only the eigenvalues and eigenvectors of A are computed if the compute_A
    flag is not True. Returns a dictionary of computed values.
    """
    # Form the leading block.
    nx = len(X)
    T1e = np.linalg.norm(X, axis=1) ** 2
    T1 = sparse.diags(T1e)

    # Form the second and third blocks.
    T2e = np.diag(X[1:].dot(X[:-1].T))
    T2 = sparse.spdiags([T2e, T2e], diags=[-1, 0], m=nx, n=nx - 1)

    # Form the final block.
    T3e = np.insert(np.diag(X[2:].dot(X[:-2].T)), 0, 0)
    T3_offdiag = sparse.spdiags(T3e, diags=1, m=nx - 1, n=nx - 1)
    T3 = (
        sparse.spdiags(T1e[:-1] + T1e[1:], diags=0, m=nx - 1, n=nx - 1)
        + T3_offdiag
        + T3_offdiag.conj().T
    )

    # Form the symmetric block-tridiagonal matrix T = [T1  T2; T2* T3].
    T = sparse.vstack(
        [sparse.hstack([T1, T2]), sparse.hstack([T2.conj().T, T3])]
    )

    # Solve for c in the system Tc = d.
    d = np.concatenate(
        (
            np.diag(X.dot(Y.T)),
            np.diag(X[:-1].dot(Y[1:].T)) + np.diag(X[1:].dot(Y[:-1].T)),
        ),
        axis=None,
    )
    c = sparse.linalg.lsqr(T.real, d.real)[0]

    # Form the solution matrix A.
    A_sparse = (
        sparse.diags(c[:nx])
        + sparse.spdiags(np.insert(c[nx:], 0, 0), diags=1, m=nx, n=nx)
        + sparse.spdiags(np.append(c[nx:], 0), diags=-1, m=nx, n=nx)
    )

    if compute_A:
        return {"A": A_sparse.toarray()}

    r = compute_rank(X, svd_rank)
    eigenvalues, modes = sparse.linalg.eigs(A_sparse, k=r)
    return {"eigenvalues": eigenvalues, "modes": modes}


def compute_BCCB(X, Y, block_shape, bccb_opt, svd_rank):
    """
    Given the data matrices X and Y and the rank truncation svd_rank, solves
    for the best-fit BCCB (block circulant with circulant blocks) matrix A that
    solves the relationship Y = AX. The blocks of A will have the shape
    block_shape and A will have the additional matrix property given by
    bccb_opt. Returns a dictionary containing A.
    """

    def fft_block2d(X):
        """
        Helper method that, given a 2D numpy array X, reshapes the
        columns of X into arrays of shape block_shape, computes the
        2D discrete Fourier Transform, and returns the resulting
        FFT matrix restored to the original X shape and scaled
        according to the length of the columns of X.
        """
        m, n = X.shape
        Y = X.reshape(*block_shape, n, order="F")
        Y_fft2 = fft2(Y, axes=(0, 1))
        return Y_fft2.reshape(m, n, order="F") / np.sqrt(m)

    nx = len(X)
    fX = fft_block2d(X.conj())
    fY = fft_block2d(Y.conj())
    d = np.zeros(nx, dtype="complex")

    if not bccb_opt:
        for j in range(nx):
            d[j] = np.dot(fX[j], fY[j].conj()).conj()
            d[j] /= np.linalg.norm(fX[j].conj(), 2) ** 2
    else:
        for j in range(nx):
            dp = np.linalg.lstsq(fX[j, :, None], fY[j, :, None], rcond=None)[0][
                0
            ][0]
            if bccb_opt == "unitary":
                d[j] = np.exp(1j * np.angle(dp))
            elif bccb_opt == "symmetric":
                d[j] = dp.real
            else:  # skewsymmetric
                d[j] = 1j * dp.imag

    # Remove the least important eigenvalues.
    r = compute_rank(X, svd_rank)
    res = np.divide(
        np.diag(np.abs(fX.dot(fY.conj().T))),
        np.linalg.norm(fX.conj().T, 2, axis=0).T,
    )
    d[np.argpartition(res, nx - r)[: nx - r]] = 0
    A = fft_block2d(np.multiply(d.conj(), fft_block2d(np.eye(nx)).conj().T).T)
    return {"A": A}


def compute_BC(X, Y, block_shape, tridiagonal_blocks=False):
    """
    Given the data matrices X and Y, solves for the best-fit block circulant
    matrix A that solves the relationship Y = AX. The blocks of A will have the
    shape block_shape and will be tridiagonal if tridiagonal_blocks is True.
    Returns a dictionary containing A.
    """

    def fft_block(X):
        """
        Helper method that, given a 2D numpy array X, reshapes the
        columns of X into arrays of shape block_shape, computes the
        1D discrete Fourier Transform, and returns the resulting
        FFT matrix restored to the original X shape and scaled
        according to the number of columns per block.
        """
        m, n = X.shape
        Y = X.reshape(*block_shape, n, order="F")
        Y_fft = fft(Y, axis=1)
        return Y_fft.reshape(m, n, order="F") / np.sqrt(block_shape[1])

    nx = len(X)
    fX = fft_block(X)
    fY = fft_block(Y)
    d = []

    for j in range(block_shape[1]):
        ls = (j * block_shape[0]) + np.arange(block_shape[0])
        if tridiagonal_blocks:
            d.append(
                compute_diagonal(
                    fX[ls], fY[ls], svd_rank=-1, manifold_opt=2, compute_A=True
                )["A"]
            )
        else:
            d.append(np.linalg.lstsq(fX[ls].T, fY[ls].T, rcond=None)[0].T)

    # Update self._A to be the full block circulant matrix.
    BD = block_diag(*d)
    fI = fft_block(np.eye(nx))
    A = fft_block(BD.dot(fI).conj()).conj()
    return {"A": A}



"""
Module for the multi-resolution Coherent Spatio-Temporal Scale Separation with
DMD.

References:
- Dylewsky, D., Tao, M., & Kutz, J. N. (2019). Dynamic mode decomposition for
multiscale nonlinear physics. Physics Review E, 99(6),
10.1103/PhysRevE.99.063311. https://doi.org/10.1103/PhysRevE.99.063311
"""

import os
import copy
import numpy as np
from sklearn.cluster import MiniBatchKMeans
from sklearn.metrics import silhouette_score, calinski_harabasz_score
import xarray as xr
from pydmd.costs import COSTS


class mrCOSTS:
    """Multi-resolution Coherent Spatio-Temporal Scale Separation (mrCOSTS)
    with DMD.

    :param window_length_array: Length of the analysis window in number of time
        steps.
    :type window_length_array: list of int
    :param step_size_array: Number of time steps to slide each CSM-DMD window.
    :type step_size_array: list of int
    :param n_components_array: Number of frequency bands to use for clustering
        each decomposition level. Only one of `cluster_sweep` or
        `n_components_array` should be provided.
    :type n_components_array: list of int
    :param svd_rank_array: The rank of the BOPDMD fit.
    :type svd_rank_array: list of int
    :param global_svd_array: Flag indicating whether to find the proj_basis and
        initial values using the entire dataset instead of individually for
        each window. Generally using the global_svd speeds up the fitting
        process by not finding a new initial value for each window. Default
        is True.
    :type cluster_sweep: bool
    :param cluster_sweep: Flag indicating whether to find the optimal value
        of `n_clusters` for the clustering of fitted eigenvalues. Only one of
        `cluster_sweep` or `n_components_array` should be provided.
    :type global_svd_array: list of bool
    :param pydmd_kwargs: Keyword arguments to pass onto the BOPDMD object.
    :type pydmd_kwargs: dict
    """

    def __init__(
        self,
        window_length_array=None,
        step_size_array=None,
        svd_rank_array=None,
        global_svd_array=None,
        pydmd_kwargs=None,
        costs_recon_kwargs=None,
        n_components_array=None,
        cluster_sweep=False,
        transform_method=None,
    ):
        self._n_components_array = n_components_array
        self._step_size_array = step_size_array
        self._window_length_array = window_length_array
        self._svd_rank_array = svd_rank_array
        self._global_svd_array = global_svd_array
        self._cluster_sweep = cluster_sweep
        self._transform_method = transform_method

        if (self._n_components_array is not None) and (
            self._step_size_array is not None
        ):
            if not len(self._n_components_array) == len(self._step_size_array):
                raise ValueError(
                    (
                        "n_components_array and step_size_array must be the"
                        "same length."
                    )
                )

        # Initialize variables that are defined in fitting.
        self._n_decompositions = None
        self._n_data_vars = None
        self._n_time_steps = None
        self._omega_classes = None
        self._costs_array = None
        self._da_omega = None
        self._n_components_global = None
        self._cluster_centroids = None

        # Specify default keywords to hand to CoSTS's BOPDMD model.
        if pydmd_kwargs is None:
            self._pydmd_kwargs = {
                "eig_sort": "imag",
                "proj_basis": None,
                "use_proj": False,
            }
        else:
            self._pydmd_kwargs = pydmd_kwargs
            self._pydmd_kwargs["eig_sort"] = pydmd_kwargs.get(
                "eig_sort", "imag"
            )
            self._pydmd_kwargs["proj_basis"] = pydmd_kwargs.get(
                "proj_basis", None
            )
            self._pydmd_kwargs["use_proj"] = pydmd_kwargs.get("use_proj", False)

        if costs_recon_kwargs is None:
            self._costs_recon_kwargs = {}
        else:
            self._costs_recon_kwargs = costs_recon_kwargs

    @property
    def costs_array(self):
        """
        :return: costs objects for each decomposition level.
        :rtype: list
        """
        return self._costs_array

    @costs_array.setter
    def costs_array(self, costs_array):
        self._costs_array = costs_array

    @property
    def svd_rank_array(self):
        """
        :return: the rank used for the svd truncation.
        :rtype: int or float
        """
        return self._svd_rank_array

    @property
    def window_length_array(self):
        """
        :return: the length of the windows used for each decomposition level.
        :rtype: list of int or float
        """
        return self._window_length_array

    @property
    def step_size_array(self):
        """
        :return: the length of the windows used for each decomposition level.
        :rtype: list of int or float
        """
        return self._step_size_array

    @property
    def n_decompositions(self):
        """
        :return: The number of multi-resolution decompositions to perform.
        :rtype: int
        """
        return self._n_decompositions

    @property
    def transform_method(self):
        """
        :return: How to transform the eigenvalues for clustering.
        :rtype: string
        """
        return self._transform_method

    @property
    def n_components_array(self):
        """
        :return: the number of frequency bands used for each decomposition
            level.
        :rtype: list of int or float
        """
        return self._n_components_array

    @property
    def cluster_centroids(self):
        """
        :return: Cluster centroids from clustering of eigenvalues.
        :rtype: list of float
        """
        if self._cluster_centroids is None:
            raise ValueError("You need to call `cluster_omega()` first.")
        return self._cluster_centroids

    @property
    def n_components_global(self):
        """
        :return: Number of global frequency bands
        :rtype: list of float
        """
        if self._n_components_global is None:
            raise ValueError(
                "You need to call `global_cluster_hyperparameter_sweep()` "
                "first or assign a the value directly."
            )
        return self._n_components_global

    # @ToDo: Use the class variable instead of passing it around
    @property
    def omega_classes_interpolated(self):
        """Returns the multi-resolution interpolation of omega classes

        :return: Ints for each omega value indicating which cluster it
            belongs to.
        :rtype: list of numpy.ndarray
        """
        if self._omega_classes is None:
            raise ValueError("You need to call `cluster_omega()` first.")
        return self._omega_classes

    @property
    def ragged_omega_classes(self):
        """Omega classes for each decomposition level after global clustering.

        :return: list of classes for each omega value for each decomposition
            level.
        :rtype: list of numpy.ndarray
        """
        if self._omega_classes is None:
            raise ValueError("You need to call `cluster_omega()` first.")
        return self.multi_res_deterp()

    @property
    def ragged_omega_array(self):
        """Omega values for each decomposition level.

        :return: list of omega arrays for each decomposition level.
        :rtype: list of numpy.ndarray
        """
        if self._costs_array is None:
            raise ValueError(
                "You need to `fit` or load previous fit from file first."
            )
        return [c.omega_array for c in self._costs_array]

    @property
    def ragged_modes_array(self):
        """Modes for each decomposition level.

        :return: list of modes arrays for each decomposition level.
        :rtype: list of numpy.ndarray
        """
        if self._costs_array is None:
            raise ValueError(
                "You need to `fit` or load previous fit from file first."
            )
        return [c.modes_array for c in self._costs_array]

    @property
    def ragged_amplitudes_array(self):
        """Amplitudes for each decomposition level.

        :return: list of amplitudes arrays for each decomposition level.
        :rtype: list of numpy.ndarray
        """
        if self._costs_array is None:
            raise ValueError(
                "You need to `fit` or load previous fit from file first."
            )
        return [c.amplitudes_array for c in self._costs_array]

    @staticmethod
    def _data_shape(data):
        """Give the data shape.

        :return: Shape of the data for fitting.
        :rtype: Tuple of ints
        """
        n_time_steps = np.shape(data)[1]
        n_data_vars = np.shape(data)[0]
        return n_time_steps, n_data_vars

    def fit(self, data, time, verbose=True):
        """
        Compute mrCOSTS for the given data.

        :param data: the input snapshots.
        :type data: numpy.ndarray
        :param time: the input time vector.
        :type time: numpy.ndarray
        :param verbose: If mrCOSTS should report progress.
        :type verbose: bool
        """
        window_lengths = self._window_length_array
        step_sizes = self._step_size_array
        svd_ranks = self._svd_rank_array
        self._n_decompositions = len(self._window_length_array)
        n_decompositions = self._n_decompositions
        transform_method = self._transform_method

        # Check for the n_components array and cluster sweeping.
        if self._cluster_sweep:
            if self._n_components_array is not None:
                raise ValueError(
                    (
                        "Only one of `cluster_sweep` and `n_components_array` "
                        "can be provided."
                    )
                )
            self._n_components_array = np.zeros(self._n_decompositions) * np.nan

        # Set the global_svd flag if none was provided.
        if self._global_svd_array is None:
            self._global_svd_array = [True] * n_decompositions

        self._costs_array = []
        self._n_time_steps, self._n_data_vars = self._data_shape(data)

        x_iter = data

        for n_decomp, (window, step, rank) in enumerate(
            zip(window_lengths, step_sizes, svd_ranks)
        ):
            global_svd = self._global_svd_array[n_decomp]

            mrd = COSTS(
                svd_rank=rank,
                global_svd=global_svd,
                pydmd_kwargs=self._pydmd_kwargs,
            )

            if verbose:
                print("_________________________________________________")
                print(f"Fitting window length = {window:}")
            mrd.fit(x_iter, np.atleast_2d(time), window, step, verbose=verbose)

            # Cluster the frequency bands
            if self._cluster_sweep or np.isnan(
                self._n_components_array[n_decomp]
            ):
                n_components = mrd.cluster_hyperparameter_sweep(
                    transform_method=transform_method
                )
                self._n_components_array[n_decomp] = n_components
            else:
                n_components = self._n_components_array[n_decomp]

            mrd.cluster_omega(
                n_components=n_components, transform_method=transform_method
            )

            # Global reconstruction
            if verbose:
                global_reconstruction = mrd.global_reconstruction(
                    scale_reconstruction_kwargs=self._costs_recon_kwargs,
                )
                re = mrd.relative_error(global_reconstruction.real, x_iter)
                print(f"Error in Global Reconstruction = {re:.2}")

            # Pass the low frequency component to the next level of
            # decomposition.
            if n_decomp < n_decompositions - 1:
                # Scale separation
                xr_low_frequency, _ = mrd.scale_separation(
                    scale_reconstruction_kwargs=self._costs_recon_kwargs
                )
                x_iter = xr_low_frequency

            # Save the fitted costs object.
            self._costs_array.append(copy.copy(mrd))

    @staticmethod
    def interp_fill(
        da_interp,
        da,
    ):
        """Extrapolate values to the full data window.

        Since multi_res_interp uses a nearest neighbors interpolation without
        exraploation and the window_time_means are in the middle of the
        decomposition window, the first and last windows are incompletely
        filled. This function fills the NaN entries of these windows.

        :param da_interp: Interpolated eigenvalues following multi_res_interp()
        :type da_interp: xarray.DataArray
        :param da: Uninterpolated eigenvalues
        :type da: xarray.DataArray
        :return: Interpolated eigenvalues with filled first and last windows.
        :rtype: xarray.DataArray
        """
        window_delta = da.window_time_means.diff(
            dim="window_time_means"
        ).values[0]

        # Backwards fill the last non-NaN to the beginning of the first window
        first_window = da_interp.dropna(
            dim="window_time_means", how="all"
        ).isel(window_time_means=0)

        da_interp = xr.where(
            (
                da_interp.window_time_means
                > da.window_time_means[0] - window_delta / 2
            )
            & (da_interp.window_time_means < da.window_time_means[0]),
            first_window,
            da_interp,
        )

        # Forward fill the last non-NaN to the end of the last window
        last_window = da_interp.dropna(dim="window_time_means", how="all").isel(
            window_time_means=-1
        )

        da_interp = xr.where(
            (
                da_interp.window_time_means
                < da.window_time_means[-1] + window_delta / 2
            )
            & (da_interp.window_time_means > da.window_time_means[-1]),
            last_window,
            da_interp,
        )

        return da_interp

    def multi_res_interp(
        self,
    ):
        """
        Interpolate the mrCOSTS eigenvalues to the smallest decomposition level.

        :return: Interpolated mrCOSTS eigenvalues
        :rtype: xarray DataArray
        """
        ds_list = [c.to_xarray() for c in self._costs_array]
        # Remove the low-frequency bands.
        da_to_concat = [
            ds_list[0].omega.where(ds_list[0].omega_classes > 0, drop=True)
        ]

        # Interpolate the larger decomposition levels to the timestep of the
        # smallest decomposition level.

        # Previously there was a mistake in the frequencies squared omega
        # transformation which made this step particularly likely to fail
        # in difficult to discover ways. The interpolation was over-engineered
        # prior to discovering this mistake, but is more explicit than the
        # previous version.
        for ds in ds_list[1:]:
            da = ds.omega.where(ds.omega_classes > 0, drop=True)
            da_real = da.real
            da_imag = da.imag

            da_real = da_real.interp(
                window_time_means=ds_list[0].window_time_means,
                method="nearest",
            )

            da_imag = da_imag.interp(
                window_time_means=ds_list[0].window_time_means,
                method="nearest",
            )

            # Complete the interpolation of the first and last window
            da_imag = self.interp_fill(da_imag, da)
            da_real = self.interp_fill(da_real, da)

            da = da_real + 1j * da_imag
            da_to_concat.append(da)

        da_omega = xr.concat(
            da_to_concat,
            dim="window_length",
        )
        da_omega.coords["window_length"] = self._window_length_array
        da_omega.coords["decomposition_level"] = (
            "window_length",
            np.arange(len(da_omega.window_length)),
        )

        self._da_omega = da_omega

    def multi_res_deterp(self):
        """
        Un-interpolate the mrCOSTS eigenvalues to the original spacing.

        :return: Omega classes from global clustering.
        :rtype: list of numpy.ndarrays
        """

        # Get the indices for the 3-d omega structure
        index = np.nonzero(~np.isnan(self._da_omega.values))

        # Unravel the flattened omega_classes. The class value of `-1` refers
        # to the slowest mode in each decomposition window which shouldn't be
        # included in the global clustering.
        omega_classes_full = (
            np.zeros_like(self._da_omega.values, dtype="int") - 1
        )
        omega_classes_full[index] = self._omega_classes

        # Build the omega_classes array into a labeled xarray DataArray object.
        da_omega_classes = xr.zeros_like(self._da_omega, dtype="int")
        da_omega_classes.values = omega_classes_full
        da_omega_classes = da_omega_classes.swap_dims(
            {"window_length": "decomposition_level"}
        )

        # Get a list of the costs results in xarray format.
        ds_list = [c.to_xarray() for c in self._costs_array]

        # Interpolate from the high-resolution of the first decomposition level
        # to the coarser time resolution of the higher decomposition levels
        omega_classes_list = [
            da_omega_classes.sel(decomposition_level=d)
            .sel(
                window_time_means=ds_list[d].window_time_means, method="nearest"
            )
            .values
            for d in da_omega_classes.decomposition_level.values
        ]

        return omega_classes_list

    def from_netcdf(self, file_list):
        """
        Create an mrCoSTS object from saved netcdf files.

        :param file_list: Filenames to open.
        :type file_list: list of str
        """
        mrd_list = []
        for f in file_list:
            if os.path.isfile(f):
                mrd_list.append(xr.load_dataset(f, engine="h5netcdf"))
            else:
                raise ValueError(f"{f:} was not found.")
        # Sort by window length
        mrd_list = sorted(mrd_list, key=lambda mrd: mrd.window_length)

        # Populate information about the fitted data.
        n_data_vars = mrd_list[0].attrs["n_data_vars"]
        n_time_steps = mrd_list[0].attrs["n_time_steps"]

        # Convert to an array of costs objects.
        mrd_list = [COSTS().from_xarray(mrd) for mrd in mrd_list]

        window_length_array = [mrd.window_length for mrd in mrd_list]
        step_size_array = [mrd.step_size for mrd in mrd_list]
        svd_rank_array = [mrd.svd_rank for mrd in mrd_list]
        n_components_array = [mrd.n_components for mrd in mrd_list]
        global_svd_array = [mrd.global_svd for mrd in mrd_list]

        # mrCOSTS currently does not support variable pydmd_kwargs
        # for each level.
        pydmd_kwargs = mrd_list[0]._pydmd_kwargs

        # Initialize the mrcosts object.
        self.__init__(
            window_length_array=window_length_array,
            step_size_array=step_size_array,
            svd_rank_array=svd_rank_array,
            global_svd_array=global_svd_array,
            pydmd_kwargs=pydmd_kwargs,
            n_components_array=n_components_array,
        )

        # Initialize variables that are defined in fitting.
        self._n_decompositions = len(mrd_list)
        self.costs_array = mrd_list
        self._n_data_vars = n_data_vars
        self._n_time_steps = n_time_steps

    def to_netcdf(self, filename, filepath="."):
        """
        Save the mrCoSTS fit to file in netcdf format.

        Each decomposition level is saved as a separate file with a common
        name and an identifier for the decomposition level.

        :param filename: Common name shared by each file.
        :type filename: str
        :param filepath: Path to save the results. Default is the current
        directory.
        :type filename: str

        """
        for c in self._costs_array:
            fname = ".".join(
                (
                    filename,
                    f"window={c.window_length:}",
                    "nc",
                )
            )
            fpath = os.path.join(filepath, fname)
            c.to_xarray().to_netcdf(
                fpath,
                engine="h5netcdf",
                invalid_netcdf=True,
            )

    def _plot_helper_data_check(self, level, data=None):
        """Checks the input data for plotting.

        :param level: Decomposition level to plot (zero indexed).
        :type level: int
        :param data: Original data, only necessary for level=0.
        :type data: numpy.ndarray
        :return: Data for plotting
        :rtype: numpy.ndarray
        """
        if data is not None:
            x_iter = data
        else:
            if level == 0:
                raise ValueError(
                    (
                        "Data must be provided when plotting the first "
                        "decomposition level"
                    )
                )
            if level > 0:
                x_iter, _ = self.costs_array[level - 1].scale_separation()

        if not x_iter.shape == (self._n_data_vars, self._n_time_steps):
            raise ValueError("Input data has the wrong shape.")

        return x_iter

    def plot_local_reconstructions(
        self,
        level,
        data=None,
        kwargs=None,
        scale_reconstruction_kwargs=None,
    ):
        """Plot reconstruction of each frequency band of a decomposition level.

        Plots are space-time diagrams assuming a 1D spatial dimension. These
        are the local frequency band clusters, not the global clusters.

        Requires the input data for the decomposition as well as
        reconstructing the fit. Deriving the input data requires providing
        the actual input data for the first decomposition. Otherwise,
        the input data is recovered by reconstructing
        decomposition = level - 1.

        :param level: Decomposition level to plot (zero indexed).
        :type level: int
        :param data: Original data, only necessary for level=0.
        :type data: numpy.ndarray
        :param kwargs: Keyword arguments given to costs.plot_reconstruction()
        :type kwargs: dict
        :param scale_reconstruction_kwargs: Arguments for reconstructing the
            fit.
        :type scale_reconstruction_kwargs: dict
        :return: None
        """

        x_iter = self._plot_helper_data_check(level, data=data)

        if kwargs is None:
            kwargs = {}

        _ = self.costs_array[level].plot_reconstructions(
            x_iter,
            scale_reconstruction_kwargs=scale_reconstruction_kwargs,
            **kwargs,
        )

    def plot_local_error(
        self,
        level,
        data=None,
        scale_reconstruction_kwargs=None,
        plot_kwargs=None,
    ):
        """Plots the error for the local decomposition fit

        Plots are a space-time diagram assuming a 1D spatial dimension.

        Determining the error requires the input data for the decomposition
        as well as reconstructing the fit. Deriving the input data requires
        providing the actual input data for the first decomposition.
        Otherwise, the input data is recovered by reconstructing
        decomposition = level - 1.

        Error is expressed as a percent.

        :param level: Decomposition level for plotting
        :type level: int
        :param data: Original data, only necessary for level=0.
        :type data: numpy.ndarray
        :param scale_reconstruction_kwargs: Arguments for reconstructing the
            fit.
        :type scale_reconstruction_kwargs: dict
        :param plot_kwargs: Arguments passed to costs.plot_error().
        :type scale_reconstruction_kwargs: dict
        :return:
        """

        x_iter = self._plot_helper_data_check(level, data=data)

        if plot_kwargs is None:
            plot_kwargs = {}

        _ = self.costs_array[level].plot_error(
            x_iter,
            scale_reconstruction_kwargs=scale_reconstruction_kwargs,
            plot_kwargs=plot_kwargs,
        )

    def plot_local_scale_separation(
        self,
        level,
        data=None,
        plot_kwargs=None,
        scale_reconstruction_kwargs=None,
    ):
        """Plot local scale separation of the high and low frequency components.

        Requires the input data for the decomposition as well as
        reconstructing the fit. Deriving the input data requires providing
        the actual input data for the first decomposition. Otherwise,
        the input data is recovered by reconstructing decomposition = level - 1.

        :param level: Decomposition level for plotting
        :type level: int
        :param data: Original data, only necessary for level=0.
        :param scale_reconstruction_kwargs: Arguments for reconstructing the
            fit.
        :param plot_kwargs: Arguments passed to costs.plot_scale_separation()
        :return fig: figure handle for the plot
        :rtype fig: matplotlib.figure()
        :return axes: matplotlib subplot instances
        :rtype fig: matplotlib.Axes()
        """

        x_iter = self._plot_helper_data_check(level, data=data)

        if plot_kwargs is None:
            plot_kwargs = {}

        fig, axes = self.costs_array[level].plot_scale_separation(
            x_iter,
            plot_kwargs=plot_kwargs,
            scale_reconstruction_kwargs=scale_reconstruction_kwargs,
        )

        return fig, axes

    def plot_local_time_series(
        self,
        space_index,
        level,
        data=None,
        scale_reconstruction_kwargs=None,
    ):
        """Plots time series for an individual point.

        Includes the input data for decomposition, the low-frequency
        component for the next decomposition level, the residual of the high
        frequency component, and the reconstructions of the frequency bands
        for the point.

        :param space_index: Index of the point in space for the 1D snapshot.
        :type space_index: int
        :param level: Decomposition level for plotting
        :type level: int
        :param data: Original data, only necessary for level=0.
        :type data: numpy.ndarray
        :param scale_reconstruction_kwargs: Arguments for reconstructing the
            fit.
        :type scale_reconstruction_kwargs: dict
        :return:
        """

        x_iter = self._plot_helper_data_check(level, data=data)

        fig, axes = self.costs_array[level].plot_time_series(
            space_index,
            x_iter,
            scale_reconstruction_kwargs=scale_reconstruction_kwargs,
        )

        return fig, axes

    def _global_cluster(
        self,
        n_components=None,
        method=MiniBatchKMeans,
        transform_method=None,
        clustering_kwargs=None,
    ):
        """Helper function for clustering global frequency bands.

        :param method: Clustering method following the sklearn pattern (has
            `fit_predict` and `n_clusters` keywords). Default is
            MiniBatchKMeans.
        :type method: method
        :param n_components: The number of clusters to find.
        :type n_components: int
        :param transform_method: How to transform omega.
        :type transform_method: str or NoneType
        :param clustering_kwargs: Arguments for clustering method.
        :type clustering_kwargs: dict
        :return cluster_centroids: Centroid values for each cluster in
            transformed omega.
        :rtype cluster_centroids: numpy.ndarray
        :return omega_classes: Global frequency band cluster identifiers.
        :rtype omega_classes: numpy.ndarray
        :return omega_array: Transformed omega with NaNs removed
        :rtype omega_array: numpy.ndarray
        """

        if self._da_omega is None:
            self.multi_res_interp()

        if transform_method is None:
            transform_method = self._transform_method

        # This step flattens the array for clustering through the numpy
        # indexing to remove NaNs. NaNs, which must be exlcuded from
        # clustering, can appear as a result of the multi-resolution
        # interpolation.
        omega_array = self._da_omega.values
        omega_array = omega_array[~np.isnan(omega_array)]
        omega_array = self.transform_omega(
            omega_array, transform_method=transform_method
        )

        if clustering_kwargs is None:
            clustering_kwargs = {}
            random_state = 0
            clustering_kwargs["random_state"] = clustering_kwargs.get(
                "random_state", random_state
            )
        clustering = method(n_clusters=n_components, **clustering_kwargs)
        if not hasattr(clustering, "fit_predict") and callable(
            getattr(clustering, "fit_predict")
        ):
            raise ValueError(
                "Clustering method must have `fit_predict()` method."
            )

        omega_classes = clustering.fit_predict(np.atleast_2d(omega_array).T)
        cluster_centroids = clustering.cluster_centers_.flatten()

        # Sort the clusters by the centroid magnitude.
        idx = np.argsort(cluster_centroids)
        lut = np.zeros_like(idx)
        lut[idx] = np.arange(n_components)
        omega_classes = lut[omega_classes]
        cluster_centroids = cluster_centroids[idx]

        return omega_classes, cluster_centroids, omega_array

    def global_cluster_hyperparameter_sweep(
        self,
        n_components_range,
        transform_method=None,
        score_method=None,
        verbose=True,
        method=MiniBatchKMeans,
        clustering_kwargs=None,
    ):
        """
        Hyperparameter search for n_components for kmeans clustering.

        :param verbose: Flag for informing the user of parameter sweep progress.
        :type verbose: bool
        :param transform_method: How to transform omega for clustering. See
            `global_cluster_omega`
        :type transform_method: str
        :param n_components_range: Values of n_components for the
            hyperparameter sweep.
        :type n_components_range: numpy.ndarray of ints
        :param method: Clustering method following the sklearn pattern (has
            `fit_predict` and `n_clusters` keywords). Default is
            MiniBatchKMeans.
        :type method: method
        :param score_method: Valid scoring methods are 'silhouette' and
            'calinski-harabasz'. Default is the silhouette score, which can
            be slow for large numbers of samples.
        :type score_method: str or None
        :param clustering_kwargs: Keywords to pass to the clustering method.
        :return score: Scores for each n_components in n_components_range
        :rtype score: numpy.ndarray
        :return n_components: Optimal n_components for frequency band separation
        :rtype n_components: int
        """

        score = np.zeros_like(n_components_range, float)

        for nind, n in enumerate(n_components_range):
            if verbose:
                print(f"fitting n_components = {n:}")
            omega_classes, _, omega = self._global_cluster(
                n_components=n,
                transform_method=transform_method,
                method=method,
                clustering_kwargs=clustering_kwargs,
            )

            if score_method is None or score_method == "silhouette":
                score[nind] = silhouette_score(
                    omega.reshape(-1, 1),
                    omega_classes.reshape(-1, 1),
                )
            # Calinski-Harabasz is not a good counter to the silhouette score
            # since it just increases with increasing number of clusters. It is
            # only included as a reference and should be replaced if a serious
            # look is given to altering the clustering algorithm.
            elif score_method == "calinski-harabasz":
                score[nind] = calinski_harabasz_score(
                    omega.reshape(-1, 1),
                    omega_classes.reshape(-1, 1),
                )

        self._n_components_global = n_components_range[np.argmax(score)]

        return score, n_components_range[np.argmax(score)]

    def global_cluster_omega(
        self,
        n_components=None,
        transform_method=None,
        clustering_kwargs=None,
        method=MiniBatchKMeans,
    ):
        """Performs frequency band clustering on the global distribution of
        omega.

        Uses k-means clustering with the MiniBatchKMeans method. The
        hyperparameter for this unsupervised method is the number of
        clusters, given by `n_components`. Transforming omega may be
        necessary to get well-separated frequency bands. Options for
        transforming omega are:
            "period": :math:`\\frac{1}{\\omega}`
            "log10": :math:`log10(\\omega)`
            "square_frequencies": :math:`\\omega^2`
            "absolute": :math:`|\\omega|`
        Default value is "absolute". All transformations and clustering are
        performed on the imaginary portion of omega.

        :param method: Clustering method following the sklearn pattern (has
            `fit_predict` and `n_clusters` keywords). Default is
            MiniBatchKMeans.
        :type method: method
        :param n_components: The number of clusters to find.
        :type n_components: int
        :param transform_method: How to transform omega.
        :type transform_method: str or NoneType
        :param clustering_kwargs: Arguments for clustering method.
        :type clustering_kwargs: dict
        :return cluster_centroids: Centroid values for each cluster in
            transformed omega.
        :rtype cluster_centroids: numpy.ndarray
        :return omega_classes: Global frequency band cluster identifiers.
        :rtype omega_classes: numpy.ndarray
        :return omega_array: Transformed omega with NaNs removed
        :rtype omega_array: numpy.ndarray
        """

        if n_components is None and self._n_components_global is None:
            raise ValueError(
                (
                    "Either perform a cluster hyperparameter sweep or provide"
                    " `n_components`"
                )
            )
        if n_components is not None:
            self._n_components_global = n_components

        omega_classes, cluster_centroids, omega_array = self._global_cluster(
            n_components=self._n_components_global,
            method=method,
            transform_method=transform_method,
            clustering_kwargs=clustering_kwargs,
        )

        self._cluster_centroids = cluster_centroids
        self._omega_classes = omega_classes

        return cluster_centroids, omega_classes, omega_array

    @staticmethod
    def transform_omega(omega_array, transform_method=None):
        """Transform omega, primarily for clustering.
        Options for transforming omega are:
            "period": :math:`\\frac{1}{\\omega}`
            "log10": :math:`log10(\\omega)`
            "square_frequencies": :math:`\\omega^2`
            "absolute": :math:`|\\omega|`
        Default value is "absolute". All transformations and clustering are
        performed on the imaginary portion of omega.

        :param omega_array: Omega values
        :type omega_array: numpy.ndarray
        :param transform_method: How to transform the imaginary component of
            omega.
        :type transform_method: str
        :return: transformed omega array
        :rtype: numpy.ndarray
        """
        # @ToDo: Move to a set-based evaluation.
        if transform_method is None or transform_method == "absolute":
            omega_array = np.abs(omega_array.imag.astype("float"))
        elif transform_method == "square_frequencies":
            omega_array = (omega_array.imag**2).real.astype("float")
        elif transform_method == "period":
            omega_array = 1 / np.abs(omega_array.imag.astype("float"))
        elif transform_method == "log10":
            omega_array = np.log10(np.abs(omega_array.imag.astype("float")))
            # Impute log10(0) with the smallest non-zero values in log10(omega).
            zero_imputer = omega_array[np.isfinite(omega_array)].min()
            omega_array[~np.isfinite(omega_array)] = zero_imputer
        else:
            # @ToDo: Return accepted methods
            raise ValueError(
                f"Transform method {transform_method} not supported."
            )

        return omega_array

    def global_scale_reconstruction(
        self,
    ):
        """Reconstruct mrCOSTS into the constituent frequency bands.

        The reconstructed data are convolved with a guassian filter since
        points near the middle of the window are more reliable than points
        at the edge of the window. Note that this will leave the beginning
        and end of time series prone to larger errors. A best practice is
        to cut off `window_length` from each end before further analysis.

        :return: Reconstruction with dimensions of:
            n_decompositions x n_components x n_data_vars x n_time_steps
        :rtype: numpy.ndarray
        """

        # Each individual reconstructed window
        xr_sep = np.zeros(
            (
                self.n_decompositions,
                self._n_components_global,
                self._n_data_vars,
                self._n_time_steps,
            )
        )

        omega_classes_list = self.multi_res_deterp()

        for n_mrd, mrd in enumerate(self._costs_array):
            # Track the total contribution from all windows to each time step
            xn = np.zeros(self._n_time_steps)

            # Convolve each windowed reconstruction with a gaussian filter.
            # Std dev of gaussian filter
            recon_filter = mrd.build_kern(mrd.window_length)

            omega_classes = omega_classes_list[n_mrd]

            if mrd.svd_rank < np.max(self._svd_rank_array):
                truncate_slice = slice(None, mrd.svd_rank)
                omega_classes = omega_classes[:, truncate_slice]

            # Iterate over each window slide performed.
            for k in range(mrd.n_slides):
                w = mrd.modes_array[k]
                b = mrd.amplitudes_array[k]
                omega = np.atleast_2d(mrd.omega_array[k]).T
                classification = omega_classes[k]

                # Compute each segment of xr starting at "t = 0"
                t = mrd.time_array[k]
                t_start = mrd.time_array[k, 0]
                t = t - t_start

                # Reconstruct each frequency band separately.
                xr_sep_window = np.zeros(
                    (
                        self._n_components_global,
                        self._n_data_vars,
                        mrd.window_length,
                    )
                )

                # Get the indices for this window.
                if k == mrd.n_slides - 1 and mrd._non_integer_n_slide:
                    # Handle non-integer number of window slides by slightly
                    # shortening the last window's slide.
                    window_indices = slice(-mrd.window_length, None)
                else:
                    window_indices = slice(
                        k * mrd.step_size,
                        k * mrd.step_size + mrd.window_length,
                    )

                for j in np.arange(0, self._n_components_global):
                    class_ind = classification == j
                    xr_sep_window[j, :, :] = np.linalg.multi_dot(
                        [
                            w[:, class_ind],
                            np.diag(b[class_ind]),
                            np.exp(omega[class_ind] * t),
                        ]
                    )

                    # Multiply by the reconstruction filter which weights
                    # the reconstruction towards the middle of the window.
                    xr_sep_window[j, :, :] = (
                        xr_sep_window[j, :, :] * recon_filter
                    )

                    xr_sep[n_mrd, j, :, window_indices] = (
                        xr_sep[n_mrd, j, :, window_indices]
                        + xr_sep_window[j, :, :]
                    )

                # A normalization factor which weights the global reconstruction
                # by the number of window centers it contains. This accounts
                # for the convolution above.
                xn[window_indices] += recon_filter

            # Normalize by the reconstruction filter.
            xr_sep[n_mrd, :, :, :] = xr_sep[n_mrd, :, :, :] / xn

        return xr_sep

    def get_background(self):
        """Build the background values not included in the scale separation.

        :return: The low frequency components of the largest decomposition
            level which are not included in the scale separation.
        :rtype: numpy.ndarray
        """

        background = self.costs_array[-1].scale_reconstruction()[0, :, :]
        return background

    def global_reconstruction(self):
        """Reconstruction using all global frequency bands and decomposition
        levels.

        :return: Global reconstruction with background component.
        :rtype: numpy.ndarray
        """
        xr_sep = self.global_scale_reconstruction()
        xr_sep = xr_sep.sum(axis=(0, 1))
        xr_background = self.get_background()

        return xr_sep + xr_background



"""Utilities module."""

import warnings
from numbers import Number
from typing import NamedTuple
from collections import namedtuple
import numpy as np
from numpy.lib.stride_tricks import sliding_window_view

#  Named tuples used in functions.
#  compute_svd uses "SVD",
#  compute_tlsq uses "TLSQ",
#  compute_rqb uses "RQB".
SVD = namedtuple("SVD", ["U", "s", "V"])
TLSQ = namedtuple("TLSQ", ["X_denoised", "Y_denoised"])
RQB = namedtuple("RQB", ["Q", "B", "test_matrix"])


def _svht(sigma_svd: np.ndarray, rows: int, cols: int) -> int:
    """
    Singular Value Hard Threshold.

    :param sigma_svd: Singual values computed by SVD
    :type sigma_svd: np.ndarray
    :param rows: Number of rows of original data matrix.
    :type rows: int
    :param cols: Number of columns of original data matrix.
    :type cols: int
    :return: Computed rank.
    :rtype: int

    References:
    Gavish, Matan, and David L. Donoho, The optimal hard threshold for
    singular values is, IEEE Transactions on Information Theory 60.8
    (2014): 5040-5053.
    https://ieeexplore.ieee.org/document/6846297
    """
    beta = np.divide(*sorted((rows, cols)))
    omega = 0.56 * beta**3 - 0.95 * beta**2 + 1.82 * beta + 1.43
    tau = np.median(sigma_svd) * omega
    rank = np.sum(sigma_svd > tau)

    if rank == 0:
        warnings.warn(
            "SVD optimal rank is 0. The largest singular values are "
            "indistinguishable from noise. Setting rank truncation to 1.",
            RuntimeWarning,
        )
        rank = 1

    return rank


def _compute_rank(
    sigma_svd: np.ndarray, rows: int, cols: int, svd_rank: Number
) -> int:
    """
    Rank computation for the truncated Singular Value Decomposition.

    :param sigma_svd: 1D singular values of SVD.
    :type sigma_svd: np.ndarray
    :param rows: Number of rows of original matrix.
    :type rows: int
    :param cols: Number of columns of original matrix.
    :type cols: int
    :param svd_rank: the rank for the truncation; If 0, the method computes
        the optimal rank and uses it for truncation; if positive interger,
        the method uses the argument for the truncation; if float between 0
        and 1, the rank is the number of the biggest singular values that
        are needed to reach the 'energy' specified by `svd_rank`; if -1,
        the method does not compute truncation. Default is 0.
    :type svd_rank: int or float
    :return: the computed rank truncation.
    :rtype: int

    References:
    Gavish, Matan, and David L. Donoho, The optimal hard threshold for
    singular values is, IEEE Transactions on Information Theory 60.8
    (2014): 5040-5053.
    """
    if svd_rank == 0:
        rank = _svht(sigma_svd, rows, cols)
    elif 0 < svd_rank < 1:
        cumulative_energy = np.cumsum(sigma_svd**2 / (sigma_svd**2).sum())
        rank = np.searchsorted(cumulative_energy, svd_rank) + 1
    elif svd_rank >= 1 and isinstance(svd_rank, int):
        rank = min(svd_rank, sigma_svd.size)
    else:
        rank = min(rows, cols)

    return rank


def compute_rank(X: np.ndarray, svd_rank: Number = 0) -> int:
    """
    Rank computation for the truncated Singular Value Decomposition.

    :param X: the matrix to decompose.
    :type X: np.ndarray
    :param svd_rank: the rank for the truncation; If 0, the method computes
        the optimal rank and uses it for truncation; if positive interger,
        the method uses the argument for the truncation; if float between 0
        and 1, the rank is the number of the biggest singular values that
        are needed to reach the 'energy' specified by `svd_rank`; if -1,
        the method does not compute truncation. Default is 0.
    :type svd_rank: int or float
    :return: the computed rank truncation.
    :rtype: int

    References:
    Gavish, Matan, and David L. Donoho, The optimal hard threshold for
    singular values is, IEEE Transactions on Information Theory 60.8
    (2014): 5040-5053.
    """
    _, s, _ = np.linalg.svd(X, full_matrices=False)
    return _compute_rank(s, X.shape[0], X.shape[1], svd_rank)


def compute_tlsq(
    X: np.ndarray, Y: np.ndarray, tlsq_rank: int
) -> NamedTuple(
    "TLSQ", [("X_denoised", np.ndarray), ("Y_denoised", np.ndarray)]
):
    """
    Compute Total Least Square.

    :param X: the first matrix;
    :type X: np.ndarray
    :param Y: the second matrix;
    :type Y: np.ndarray
    :param tlsq_rank: the rank for the truncation; If 0, the method
        does not compute any noise reduction; if positive number, the
        method uses the argument for the SVD truncation used in the TLSQ
        method.
    :type tlsq_rank: int
    :return: the denoised matrix X, the denoised matrix Y
    :rtype: NamedTuple("TLSQ", [('X_denoised', np.ndarray),
                                ('Y_denoised', np.ndarray)])

    References:
    https://arxiv.org/pdf/1703.11004.pdf
    https://arxiv.org/pdf/1502.03854.pdf
    """
    # Do not perform tlsq
    if tlsq_rank == 0:
        return X, Y

    V = np.linalg.svd(np.append(X, Y, axis=0), full_matrices=False)[-1]
    rank = min(tlsq_rank, V.shape[0])
    VV = V[:rank, :].conj().T.dot(V[:rank, :])

    return TLSQ(X.dot(VV), Y.dot(VV))


def compute_svd(
    X: np.ndarray, svd_rank: Number = 0
) -> NamedTuple(
    "SVD", [("U", np.ndarray), ("s", np.ndarray), ("V", np.ndarray)]
):
    """
    Truncated Singular Value Decomposition.

    :param X: the matrix to decompose.
    :type X: np.ndarray
    :param svd_rank: the rank for the truncation; If 0, the method computes
        the optimal rank and uses it for truncation; if positive interger,
        the method uses the argument for the truncation; if float between 0
        and 1, the rank is the number of the biggest singular values that
        are needed to reach the 'energy' specified by `svd_rank`; if -1,
        the method does not compute truncation. Default is 0.
    :type svd_rank: int or float
    :return: the truncated left-singular vectors matrix, the truncated
        singular values array, the truncated right-singular vectors matrix.
    :rtype: NamedTuple("SVD", [('U', np.ndarray),
                               ('s', np.ndarray),
                               ('V', np.ndarray)])

    References:
    Gavish, Matan, and David L. Donoho, The optimal hard threshold for
    singular values is, IEEE Transactions on Information Theory 60.8
    (2014): 5040-5053.
    """
    U, s, V = np.linalg.svd(X, full_matrices=False)
    rank = _compute_rank(s, X.shape[0], X.shape[1], svd_rank)
    V = V.conj().T

    U = U[:, :rank]
    V = V[:, :rank]
    s = s[:rank]

    return SVD(U, s, V)


def compute_rqb(
    X: np.ndarray,
    svd_rank: Number,
    oversampling: int,
    power_iters: int,
    test_matrix: np.ndarray = None,
    seed: int = None,
) -> NamedTuple(
    "RQB", [("Q", np.ndarray), ("B", np.ndarray), ("test_matrix", np.ndarray)]
):
    """
    Randomized QB Decomposition.

    :param X: the matrix to decompose.
    :type X: np.ndarray
    :param svd_rank: the rank for the truncation; If 0, the method computes
        the optimal rank and uses it for truncation; if positive interger,
        the method uses the argument for the truncation; if float between 0
        and 1, the rank is the number of the biggest singular values that
        are needed to reach the 'energy' specified by `svd_rank`; if -1,
        the method does not compute truncation. Use this parameter to
        define the target rank of the input matrix.
    :type svd_rank: int or float
    :param oversampling: Number of additional samples (beyond the target rank)
        to use when computing the random test matrix. Note that values in the
        range [5, 10] tend to be sufficient.
    :type oversampling: int
    :param power_iters: Number of power iterations to perform when executing
        the Randomized QB Decomposition. Note that as many as 1 to 2 power
        iterations often lead to considerable improvements.
    :type power_iters: int
    :param test_matrix: The random test matrix that will be used when executing
        the Randomized QB Decomposition. If not provided, the `svd_rank` and
        `oversampling` parameters will be used to compute the random matrix.
    :type test_matrix: numpy.ndarray
    :param seed: Seed used to initialize the random generator when computing
        random test matrices.
    :type seed: int
    :return: the orthonormal basis matrix, the transformed data matrix, and
        the random test matrix.
    :rtype: NamedTuple("RQB", [('Q', np.ndarray),
                               ('B', np.ndarray),
                               ('test_matrix', np.ndarray)])

    References:
    N. Benjamin Erichson, Lionel Mathelin, J. Nathan Kutz, Steven L. Brunton.
    Randomized dynamic mode decomposition. SIAM Journal on Applied Dynamical
    Systems, 18, 2019.
    """
    if X.ndim != 2:
        raise ValueError("Please ensure that input data is a 2D array.")

    # Define the random test matrix if not provided.
    if test_matrix is None:
        m = X.shape[-1]
        r = compute_rank(X, svd_rank)
        rng = np.random.default_rng(seed)
        test_matrix = rng.standard_normal((m, r + oversampling))

    # Compute sampling matrix.
    Y = X.dot(test_matrix)

    # Perform power iterations.
    for _ in range(power_iters):
        Q = np.linalg.qr(Y)[0]
        Z = np.linalg.qr(X.conj().T.dot(Q))[0]
        Y = X.dot(Z)

    # Orthonormalize the sampling matrix.
    Q = np.linalg.qr(Y)[0]

    # Project the snapshot matrix onto the smaller space.
    B = Q.conj().T.dot(X)

    return RQB(Q, B, test_matrix)


def pseudo_hankel_matrix(X: np.ndarray, d: int) -> np.ndarray:
    """
    Arrange the snapshot in the matrix `X` into the (pseudo) Hankel
    matrix. The attribute `d` controls the number of snapshot from `X` in
    each snapshot of the Hankel matrix.

    :Example:

        >>> a = np.array([[1, 2, 3, 4, 5]])
        >>> pseudo_hankel_matrix(a, d=2)
        array([[1, 2, 3, 4],
               [2, 3, 4, 5]])
        >>> pseudo_hankel_matrix(a, d=4)
        array([[1, 2],
               [2, 3],
               [3, 4],
               [4, 5]])

        >>> a = np.array([1,2,3,4,5,6]).reshape(2,3)
        >>> print(a)
        [[1 2 3]
         [4 5 6]]
        >>> pseudo_hankel_matrix(a, d=2)
        array([[1, 2],
               [4, 5],
               [2, 3],
               [5, 6]])
    """
    return (
        sliding_window_view(X.T, (d, X.shape[0]))[:, 0]
        .reshape(X.shape[1] - d + 1, -1)
        .T
    )


def differentiate(X, dt):
    """
    Method for performing 2nd order finite difference. Assumes that the
    input matrix X is 2D, with uniformly-sampled snapshots filling each
    column. Requires dt, which is the time step between each snapshot.
    """
    if not isinstance(X, np.ndarray) or X.ndim > 2:
        raise ValueError("Please ensure that input data is a 1D or 2D array.")
    if X.ndim == 1:
        X = X[None]
    X_prime = np.empty(X.shape)
    X_prime[:, 1:-1] = (X[:, 2:] - X[:, :-2]) / (2 * dt)
    X_prime[:, 0] = (-3 * X[:, 0] + 4 * X[:, 1] - X[:, 2]) / (2 * dt)
    X_prime[:, -1] = (3 * X[:, -1] - 4 * X[:, -2] + X[:, -3]) / (2 * dt)
    return np.squeeze(X_prime)



import numpy as np
import matplotlib.pyplot as plt
from pytest import raises, warns
from scipy.integrate import solve_ivp

from pydmd.bopdmd import BOPDMD


def simulate_z(t_eval):
    """
    Given a time vector t_eval = t1, t2, ..., evaluates and returns
    the snapshots z(t1), z(t2), ... as columns of the matrix Z.
    Simulates data z given by the system of ODEs
        z' = Az
    where A = [1 -2; 1 -1] and z_0 = [1, 0.1].
    """

    def ode_sys(zt, z):
        z1, z2 = z
        return [z1 - 2 * z2, z1 - z2]

    # Set integrator keywords to replicate odeint defaults.
    integrator_keywords = {}
    integrator_keywords["rtol"] = 1e-12
    integrator_keywords["method"] = "LSODA"
    integrator_keywords["atol"] = 1e-12

    sol = solve_ivp(
        ode_sys,
        [t_eval[0], t_eval[-1]],
        [1.0, 0.1],
        t_eval=t_eval,
        **integrator_keywords,
    )

    return sol.y


def sort_imag(x):
    """
    Helper method that sorts the entries of x by imaginary component, and then
    by real component in order to break ties.
    """
    x_real_imag_swapped = x.imag + 1j * x.real
    sorted_inds = np.argsort(x_real_imag_swapped)
    return x[sorted_inds]


def compute_error(actual, truth):
    """
    Helper method that computes relative error.
    """
    return np.linalg.norm(truth - actual) / np.linalg.norm(truth)


# Simulate data.
t = np.arange(4000) * 0.01
t_long = np.arange(6000) * 0.01
t_uneven = np.delete(t, np.arange(1000)[1::2])
Z = simulate_z(t)
Z_long = simulate_z(t_long)
Z_uneven = np.delete(Z, np.arange(1000)[1::2], axis=1)
rng = np.random.default_rng(seed=1234)
Z_noisy = Z + 0.01 * rng.standard_normal(Z.shape)

# Define the true eigenvalues of the system.
expected_eigs = np.array((-1j, 1j))

# Define the true system operator.
expected_A = np.array(((1, -2), (1, -1)))


def test_truncation_shape():
    """
    Tests that, when given a positive integer rank truncation, the shape of the
    modes, eigenvalues, amplitudes, Atilde operator, and A matrix are accurate.
    """
    bopdmd = BOPDMD(svd_rank=2, compute_A=True)
    bopdmd.fit(Z, t)
    assert bopdmd.modes.shape[1] == 2
    assert len(bopdmd.eigs) == 2
    assert len(bopdmd.amplitudes) == 2
    assert bopdmd.atilde.shape == (2, 2)
    assert bopdmd.A.shape == (2, 2)


def test_eigs():
    """
    Tests that the computed eigenvalues are accurate for the following cases:
    - standard optimized dmd, even dataset
    - standard optimized dmd, uneven dataset
    - standard optimized dmd, fit full data
    - optimized dmd with bagging
    """
    bopdmd = BOPDMD(svd_rank=2)
    bopdmd.fit(Z, t)
    np.testing.assert_allclose(sort_imag(bopdmd.eigs), expected_eigs)

    bopdmd = BOPDMD(svd_rank=2)
    bopdmd.fit(Z_uneven, t_uneven)
    np.testing.assert_allclose(sort_imag(bopdmd.eigs), expected_eigs)

    bopdmd = BOPDMD(svd_rank=2, use_proj=False)
    bopdmd.fit(Z, t)
    np.testing.assert_allclose(sort_imag(bopdmd.eigs), expected_eigs)

    bopdmd = BOPDMD(svd_rank=2, num_trials=10, trial_size=0.8)
    bopdmd.fit(Z, t)
    np.testing.assert_allclose(sort_imag(bopdmd.eigs), expected_eigs)


def test_A():
    """
    Tests that the computed A matrix is accurate for the following cases:
    - standard optimized dmd, even dataset
    - standard optimized dmd, uneven dataset
    - standard optimized dmd, noisy dataset
    - standard optimized dmd, fit full data
    - optimized dmd with bagging
    """
    bopdmd = BOPDMD(svd_rank=2, compute_A=True)
    bopdmd.fit(Z, t)
    assert compute_error(bopdmd.A, expected_A) < 1e-3

    bopdmd = BOPDMD(svd_rank=2, compute_A=True)
    bopdmd.fit(Z_uneven, t_uneven)
    assert compute_error(bopdmd.A, expected_A) < 1e-3

    bopdmd = BOPDMD(svd_rank=2, compute_A=True)
    bopdmd.fit(Z_noisy, t)
    assert compute_error(bopdmd.A, expected_A) < 1e-3

    bopdmd = BOPDMD(svd_rank=2, compute_A=True, use_proj=False)
    bopdmd.fit(Z, t)
    assert compute_error(bopdmd.A, expected_A) < 1e-3

    bopdmd = BOPDMD(svd_rank=2, compute_A=True, num_trials=10, trial_size=0.8)
    bopdmd.fit(Z, t)
    assert compute_error(bopdmd.A, expected_A) < 1e-3


def test_reconstruction():
    """
    Tests the accuracy of the reconstructed data.
    Tests for both standard optimized dmd and BOP-DMD.
    """
    bopdmd = BOPDMD(svd_rank=2)
    bopdmd.fit(Z, t)
    np.testing.assert_allclose(bopdmd.reconstructed_data, Z, rtol=1e-5)

    bopdmd = BOPDMD(svd_rank=2, num_trials=10, trial_size=0.8)
    bopdmd.fit(Z, t)
    np.testing.assert_allclose(bopdmd.reconstructed_data, Z, rtol=1e-5)


def test_forecast():
    """
    Tests that the BOP-DMD model generalizes to data not used for training.
    Tests the following cases:
    - Generalizing to long dataset after training on short, even dataset
        without bagging.
    - Generalizing to long dataset after training on short, uneven dataset
        without bagging.
    - Generalizing to long dataset after training on short, even dataset
        with bagging.
    """
    bopdmd = BOPDMD(svd_rank=2)
    bopdmd.fit(Z, t)
    assert compute_error(bopdmd.forecast(t_long), Z_long) < 1e-2

    bopdmd = BOPDMD(svd_rank=2)
    bopdmd.fit(Z_uneven, t_uneven)
    assert compute_error(bopdmd.forecast(t_long), Z_long) < 1e-2

    bopdmd = BOPDMD(svd_rank=2, num_trials=10, trial_size=0.8)
    bopdmd.fit(Z, t)
    assert compute_error(bopdmd.forecast(t_long)[0], Z_long) < 1e-2


def test_compute_A():
    """
    Tests that the BOPDMD module appropriately calculates or doesn't calculate
    A depending on the compute_A flag. Also tests that atilde, the dmd modes,
    and the dmd eigenvalues are not effected by the compute_A flag.
    """
    bopdmd_with_A = BOPDMD(svd_rank=2, compute_A=True)
    bopdmd_no_A = BOPDMD(svd_rank=2, compute_A=False)
    bopdmd_with_A.fit(Z, t)
    bopdmd_no_A.fit(Z, t)

    np.testing.assert_allclose(bopdmd_with_A.A, expected_A)
    with raises(ValueError):
        print(bopdmd_no_A.A)

    np.testing.assert_array_equal(bopdmd_with_A.atilde, bopdmd_no_A.atilde)
    np.testing.assert_array_equal(bopdmd_with_A.modes, bopdmd_no_A.modes)
    np.testing.assert_array_equal(bopdmd_with_A.eigs, bopdmd_no_A.eigs)


def test_eig_constraints_errors():
    """
    Tests that the BOPDMD module correctly throws an error upon initialization
    in each of the following cases:
    - eig_constraints is a string rather than a set of strings
    - eig_constraints contains an invalid constraint argument
        (either alone or along with another constraint argument)
    - eig_constraints contains the invalid combination "stable"+"imag"
        (either alone or along with the extra argument "conjugate_pairs")
    """
    with raises(TypeError):
        BOPDMD(eig_constraints="stable")

    with raises(ValueError):
        BOPDMD(eig_constraints={"stablee"})

    with raises(ValueError):
        BOPDMD(eig_constraints={"stablee", "imag"})

    with raises(ValueError):
        BOPDMD(eig_constraints={"stable", "imag"})

    with raises(ValueError):
        BOPDMD(eig_constraints={"stable", "imag", "conjugate_pairs"})


def test_eig_constraints():
    """
    Tests that the BOPDMD module correctly enforces that eigenvalues...
    - lie in the left half plane when eig_constraints contains "stable".
    - lie on the imaginary axis when eig_constraints contains "imag".
    - are present with their complex conjugate when eig_constraints
        contains "conjugate_pairs".
    Eigenvalue constraint combinations are also tested.
    """
    bopdmd = BOPDMD(svd_rank=2, eig_constraints={"stable"})
    bopdmd.fit(Z, t)
    assert np.all(bopdmd.eigs.real <= 0.0)

    bopdmd = BOPDMD(svd_rank=2, eig_constraints={"imag"})
    bopdmd.fit(Z, t)
    assert np.all(bopdmd.eigs.real == 0.0)

    bopdmd = BOPDMD(svd_rank=2, eig_constraints={"conjugate_pairs"})
    bopdmd.fit(Z, t)
    assert bopdmd.eigs[0].real == bopdmd.eigs[1].real
    assert bopdmd.eigs[0].imag == -bopdmd.eigs[1].imag

    bopdmd = BOPDMD(svd_rank=2, eig_constraints={"stable", "conjugate_pairs"})
    bopdmd.fit(Z, t)
    assert np.all(bopdmd.eigs.real <= 0.0)
    assert bopdmd.eigs[0].real == bopdmd.eigs[1].real
    assert bopdmd.eigs[0].imag == -bopdmd.eigs[1].imag

    bopdmd = BOPDMD(svd_rank=2, eig_constraints={"imag", "conjugate_pairs"})
    bopdmd.fit(Z, t)
    assert np.all(bopdmd.eigs.real == 0.0)
    assert bopdmd.eigs[0].imag == -bopdmd.eigs[1].imag


def test_eig_constraints_errors_2():
    """
    Tests that the BOPDMD module correctly throws an error upon initialization
    whenever eig_constraints is a function and...
    - eig_constraints is incompatible with general (n,) numpy.ndarray inputs
    - eig_constraints doesn't return a single numpy.ndarray
    - eig_constraints takes multiple arguments
    - eig_constraints changes the shape of the input array
    """

    # Function that assumes the input is length 3.
    def bad_func_1(x):
        return np.multiply(x, np.arange(3))

    # Function that assumes the input array is at least 2-dimensional.
    def bad_func_2(x):
        return x[0, :] + x[1, :]

    # Function that doesn't return an array.
    def bad_func_3(x):
        return len(x)

    # Function that returns 2 arrays instead of 1.
    def bad_func_4(x):
        return x, 2 * x

    # Function that accepts more than 1 input.
    def bad_func_5(x, y):
        return x + y

    # Function that returns an array with a different shape.
    def bad_func_6(x):
        return x[:-1]

    # Function that returns an array with an extra dimension.
    def bad_func_7(x):
        return x[:, None]

    with raises(ValueError):
        BOPDMD(eig_constraints=bad_func_1)

    with raises(ValueError):
        BOPDMD(eig_constraints=bad_func_2)

    with raises(ValueError):
        BOPDMD(eig_constraints=bad_func_3)

    with raises(ValueError):
        BOPDMD(eig_constraints=bad_func_4)

    with raises(ValueError):
        BOPDMD(eig_constraints=bad_func_5)

    with raises(ValueError):
        BOPDMD(eig_constraints=bad_func_6)

    with raises(ValueError):
        BOPDMD(eig_constraints=bad_func_7)

    # See that bad_func_1 is fine if the svd_rank is 3.
    BOPDMD(svd_rank=3, eig_constraints=bad_func_1)


def test_eig_constraints_2():
    """
    Tests that if the eig_constraints function discards all real parts,
    the functionality is the same as setting eig_constraints={"imag"}.
    """

    def make_imag(x):
        return 1j * x.imag

    bopdmd1 = BOPDMD(svd_rank=2, eig_constraints={"imag"}).fit(Z, t)
    bopdmd2 = BOPDMD(svd_rank=2, eig_constraints=make_imag).fit(Z, t)
    np.testing.assert_array_equal(bopdmd1.eigs, bopdmd2.eigs)


def test_plot_mode_uq():
    """
    Test that a basic call to plot_mode_uq is successful.
    """
    bopdmd = BOPDMD(svd_rank=2, num_trials=10, trial_size=0.8)
    bopdmd.fit(Z, t)
    bopdmd.plot_mode_uq()
    plt.close()


def test_plot_eig_uq():
    """
    Test that a basic call to plot_eig_uq is successful.
    """
    bopdmd = BOPDMD(svd_rank=2, num_trials=10, trial_size=0.8)
    bopdmd.fit(Z, t)
    bopdmd.plot_eig_uq()
    plt.close()


def test_plot_error():
    """
    Test that UQ plotters fail if bagging wasn't used.
    """
    bopdmd = BOPDMD(svd_rank=2, num_trials=0)
    bopdmd.fit(Z, t)

    with raises(ValueError):
        bopdmd.plot_mode_uq()

    with raises(ValueError):
        bopdmd.plot_eig_uq()


def test_varpro_opts_warn():
    """
    Test that errors or warnings are correctly thrown if invalid
    or poorly-chosen variable projection parameters are given.
    The `tol` parameter is specifically tested.
    """
    with raises(TypeError):
        bopdmd = BOPDMD(varpro_opts_dict={"tol": None})
        bopdmd.fit(Z, t)

    with warns():
        bopdmd = BOPDMD(varpro_opts_dict={"tol": -1.0})
        bopdmd.fit(Z, t)

    with warns():
        bopdmd = BOPDMD(varpro_opts_dict={"tol": np.inf})
        bopdmd.fit(Z, t)


def test_varpro_opts_print():
    """
    Test that variable projection parameters can be printed after fitting.
    """
    bopdmd = BOPDMD(svd_rank=2)

    with raises(ValueError):
        bopdmd.print_varpro_opts()

    bopdmd.fit(Z, t)
    bopdmd.print_varpro_opts()


def test_verbose_outputs_1():
    """
    Test variable projection verbosity for optimized DMD.
    """
    bopdmd = BOPDMD(
        svd_rank=2,
        num_trials=0,
        varpro_opts_dict={"verbose": True},
    )
    bopdmd.fit(Z, t)


def test_verbose_outputs_2():
    """
    Test variable projection verbosity for BOP-DMD.
    """
    bopdmd = BOPDMD(
        svd_rank=2,
        num_trials=10,
        trial_size=0.8,
        varpro_opts_dict={"verbose": True},
    )
    bopdmd.fit(Z, t)


def test_verbose_outputs_3():
    """
    Test variable projection verbosity for BOP-DMD without bad bags.
    """
    bopdmd = BOPDMD(
        svd_rank=2,
        num_trials=10,
        trial_size=0.8,
        varpro_opts_dict={"verbose": True, "tol": 1.0},
        remove_bad_bags=True,
    )
    bopdmd.fit(Z, t)


def test_bag_int():
    """
    Test that trial_size can be a valid integer value.
    """
    bopdmd = BOPDMD(svd_rank=2, num_trials=10, trial_size=3200)
    bopdmd.fit(Z, t)


def test_bag_getters():
    """
    Test calls to the num_trials and trial_size parameters.
    """
    bopdmd = BOPDMD(svd_rank=2, num_trials=0, trial_size=0.8)
    assert bopdmd.num_trials == 0
    assert bopdmd.trial_size == 0.8

    bopdmd = BOPDMD(svd_rank=2, num_trials=100, trial_size=3200)
    assert bopdmd.num_trials == 100
    assert bopdmd.trial_size == 3200


def test_bag_error():
    """
    Test that errors are thrown if invalid bagging parameters are given.
    """
    # Error should raise if trial_size isn't a positive integer...
    with raises(ValueError):
        bopdmd = BOPDMD(svd_rank=2, num_trials=10, trial_size=-1)
        bopdmd.fit(Z, t)

    # ...or if it isn't a float in the range (0.0, 1.0).
    with raises(ValueError):
        bopdmd = BOPDMD(svd_rank=2, num_trials=10, trial_size=2.0)
        bopdmd.fit(Z, t)

    # Error should raise if the requested trial size is too big...
    with raises(ValueError):
        bopdmd = BOPDMD(svd_rank=2, num_trials=10, trial_size=4001)
        bopdmd.fit(Z, t)

    # ...or if the requested trial size is too small.
    with raises(ValueError):
        bopdmd = BOPDMD(svd_rank=2, num_trials=10, trial_size=1e-6)
        bopdmd.fit(Z, t)


def test_mode_prox():
    """
    Test that the mode_prox function is applied as expected.
    """

    def dummy_prox(X):
        return X + 1.0

    # Test that the function is applied in the use_proj=False case.
    bopdmd = BOPDMD(svd_rank=2, use_proj=False, mode_prox=dummy_prox)
    bopdmd.fit(Z, t)

    # Test that the function is applied in the use_proj=True case.
    bopdmd = BOPDMD(svd_rank=2, use_proj=True, mode_prox=dummy_prox)
    bopdmd.fit(Z, t)

    # Compare use_proj=True results to the no prox case.
    bopdmd_noprox = BOPDMD(svd_rank=2, use_proj=True)
    bopdmd_noprox.fit(Z, t)
    np.testing.assert_allclose(bopdmd.modes, dummy_prox(bopdmd_noprox.modes))


def test_init_alpha_initializer():
    """
    Test that the eigenvalues are accurately initialized by default.
    """
    bopdmd = BOPDMD(svd_rank=2)

    # Initial eigs shouldn't be defined yet.
    with raises(ValueError):
        _ = bopdmd.init_alpha

    # After fitting, the initial eigs used should be fairly accurate.
    bopdmd.fit(Z, t)
    np.testing.assert_allclose(
        sort_imag(bopdmd.init_alpha),
        expected_eigs,
        rtol=0.01,
    )


def test_proj_basis_initializer():
    """
    Test that the projection basis is accurately initialized by default.
    """
    bopdmd = BOPDMD(svd_rank=2)

    # Projection basis shouldn't be defined yet.
    with raises(ValueError):
        _ = bopdmd.proj_basis

    # After fitting, the projection basis used should be accurate.
    bopdmd.fit(Z, t)
    np.testing.assert_array_equal(bopdmd.proj_basis, np.linalg.svd(Z)[0])


def test_svd_rank():
    """
    Test svd_rank getter and setter.
    """
    bopdmd = BOPDMD(svd_rank=2)
    assert bopdmd.svd_rank == 2

    bopdmd.svd_rank = 0
    assert bopdmd.svd_rank == 0


def test_init_alpha():
    """
    Test init_alpha getter and setter.
    """
    dummy_eigs = 2.0 * expected_eigs
    bopdmd = BOPDMD(init_alpha=dummy_eigs)
    np.testing.assert_array_equal(bopdmd.init_alpha, dummy_eigs)

    bopdmd.init_alpha = 2.0 * dummy_eigs
    np.testing.assert_array_equal(bopdmd.init_alpha, 2.0 * dummy_eigs)


def test_proj_basis():
    """
    Test proj_basis getter and setter.
    """
    dummy_basis = 2.0 * np.linalg.svd(Z)[0]
    bopdmd = BOPDMD(proj_basis=dummy_basis)
    np.testing.assert_array_equal(bopdmd.proj_basis, dummy_basis)

    bopdmd.proj_basis = 2.0 * dummy_basis
    np.testing.assert_array_equal(bopdmd.proj_basis, 2.0 * dummy_basis)


def test_default_initializers():
    """
    Test that default initialization does not impact custom inputs.
    """
    # Define the dummy init_alpha and proj_basis to NOT be the defaults.
    dummy_eigs = 2.0 * expected_eigs
    dummy_basis = 2.0 * np.linalg.svd(Z)[0]
    bopdmd = BOPDMD(
        svd_rank=2,
        init_alpha=dummy_eigs,
        proj_basis=dummy_basis,
    )
    bopdmd.fit(Z, t)
    np.testing.assert_array_equal(bopdmd.init_alpha, dummy_eigs)
    np.testing.assert_array_equal(bopdmd.proj_basis, dummy_basis)


def test_std_shape():
    """
    Test the shapes of the standard deviation attributes.
    """
    bopdmd = BOPDMD(svd_rank=2, num_trials=10, trial_size=0.8)
    bopdmd.fit(Z, t)

    assert bopdmd.eigenvalues_std.shape == bopdmd.eigs.shape
    assert bopdmd.modes_std.shape == bopdmd.modes.shape
    assert bopdmd.amplitudes_std.shape == bopdmd.amplitudes.shape


def test_std_nobags():
    """
    Test that std attributes are simply None if no bags were used.
    """
    bopdmd = BOPDMD(svd_rank=2, num_trials=0)
    bopdmd.fit(Z, t)

    assert bopdmd.eigenvalues_std is None
    assert bopdmd.modes_std is None
    assert bopdmd.amplitudes_std is None


def test_getter_errors():
    """
    Test that error occurs if the following properties are called before fit:
    time, atilde, A, eigenvalues_std, modes_std, amplitudes_std.
    """
    bopdmd = BOPDMD(compute_A=True)

    with raises(ValueError):
        _ = bopdmd.time

    with raises(ValueError):
        _ = bopdmd.atilde

    with raises(ValueError):
        _ = bopdmd.A

    with raises(ValueError):
        _ = bopdmd.eigenvalues_std

    with raises(ValueError):
        _ = bopdmd.modes_std

    with raises(ValueError):
        _ = bopdmd.amplitudes_std



"""
Derived module from dmdbase.py for linear and
nonlinear disambiguation optimization (LANDO).

References:
- Baddoo Peter J., Herrmann Benjamin, McKeon Beverley J. and Brunton Steven L.
2022 Kernel learning for robust dynamic mode decomposition: linear and
nonlinear disambiguation optimization. Proc. R. Soc. A. 478: 20210830.
20210830. http://doi.org/10.1098/rspa.2021.0830
"""

import warnings
from inspect import isfunction

import numpy as np
from scipy.integrate import solve_ivp
from sklearn.metrics.pairwise import pairwise_kernels

from .dmdbase import DMDBase
from .dmdoperator import DMDOperator
from .snapshots import Snapshots
from .utils import compute_svd, compute_tlsq


class LANDOOperator(DMDOperator):
    """
    LANDO operator class, which is used to compute and keep track of the
    dictionary-based kernel model and the diagnostics of the linear model
    at a fixed point. See LANDO documentation for parameter descriptions.
    """

    def __init__(self, svd_rank, dict_tol, online, permute, lstsq):
        super().__init__(
            svd_rank=svd_rank,
            exact=True,
            forward_backward=False,
            rescale_mode=None,
            sorted_eigs=False,
            tikhonov_regularization=None,
        )

        # Keep track of the sparse dictionary.
        self._sparse_dictionary = None
        self._dict_tol = dict_tol
        self._online = online
        self._permute = permute
        self._lstsq = lstsq

        # Keep track of attributes for online learning.
        self._cholesky = None
        self._P = None

        # Keep track of operator attributes.
        self._weights = None
        self._eigenvalues = None
        self._eigenvectors = None
        self._modes = None
        self._Atilde = None
        self._A = None

    @property
    def sparse_dictionary(self):
        """
        Get the learned sparse feature dictionary.

        :return: the sparse feature dictionary.
        :rtype: numpy.ndarray
        """
        if self._sparse_dictionary is None:
            raise ValueError("You need to call fit().")
        return self._sparse_dictionary

    @property
    def weights(self):
        """
        Get the dictionary-based kernel model weights.
        Referred to as W_tilde in the original reference.

        :return: the kernel model weights.
        :rtype: numpy.ndarray
        """
        if self._weights is None:
            raise ValueError("You need to call fit().")
        return self._weights

    @property
    def eigenvalues(self):
        """
        Get the eigenvalues of the full linear operator evaluated at a fixed
        point. This operator is referred to as L in the original reference.

        :return: the eigenvalues of the full linear operator.
        :rtype: numpy.ndarray
        """
        if self._eigenvalues is None:
            msg = "You need to call fit() and analyze_fixed_point()."
            raise ValueError(msg)
        return self._eigenvalues

    @property
    def eigenvectors(self):
        """
        Get the eigenvectors of the reduced linear operator evaluated at a
        fixed point. The reduced operator is referred to as L_hat in the
        original reference.

        :return: the eigenvectors of the reduced linear operator.
        :rtype: numpy.ndarray
        """
        if self._eigenvectors is None:
            msg = "You need to call fit() and analyze_fixed_point()."
            raise ValueError(msg)
        return self._eigenvectors

    @property
    def modes(self):
        """
        Get the eigenvectors of the full linear operator evaluated at a
        fixed point. The full operator is referred to as L in the original
        reference.

        :return: the eigenvectors of the full linear operator.
        :rtype: numpy.ndarray
        """
        if self._modes is None:
            msg = "You need to call fit() and analyze_fixed_point()."
            raise ValueError(msg)
        return self._modes

    @property
    def as_numpy_array(self):
        """
        Get the reduced linear operator A_tilde.
        Referred to as L_hat in the original reference.

        :return: the reduced linear operator.
        :rtype: numpy.ndarray
        """
        if self._Atilde is None:
            msg = "You need to call fit() and analyze_fixed_point()."
            raise ValueError(msg)
        return self._Atilde

    @property
    def A(self):
        """
        Get the full linear operator A.
        Referred to as L in the original reference.

        :return: the full linear operator.
        :rtype: numpy.ndarray
        """
        if self._A is None:
            msg = (
                "Full linear operator currently not computed. You may need to "
                "call fit() and analyze_fixed_point(). If already done so, "
                "call analyze_fixed_point() using the flag compute_A=True."
            )
            warnings.warn(msg)
        return self._A

    def compute_operator(self, X, Y, kernel_function, updating=False):
        """
        Compute the dictionary-based kernel model weights.

        :param X: the input snapshots x, rescaled.
        :type X: numpy.ndarray
        :param Y: input snapshots y such that F(x) = y.
        :type Y: numpy.ndarray
        :param kernel_function: kernel function to apply.
        :type kernel_function: function
        :param updating: whether or not the operator is
            being updated after a previous fit.
        :type updating: bool
        """
        if updating and not self._online:
            msg = "Only LANDO models fitted with online=True can be updated."
            raise ValueError(msg)

        # Determine the order in which to parse the data snapshots.
        parsing_inds = np.arange(X.shape[1])
        if self._permute:
            np.random.shuffle(parsing_inds)

        # Initialize values if fitting for the first time.
        if not updating:
            # Initialize the Cholesky factorization routine.
            ind_0 = parsing_inds[0]
            x_0 = X[:, ind_0][..., None]
            y_0 = Y[:, ind_0][..., None]
            C = np.sqrt(kernel_function(x_0, x_0))
            self._sparse_dictionary = np.copy(x_0)
            parsing_inds = parsing_inds[1:]

            # Initialize the online learning routine, if applicable.
            if self._online:
                self._cholesky = C
                self._P = np.ones((1, 1))
                self._weights = y_0 / (self._cholesky[0][0] ** 2)

        for ind_t in parsing_inds:
            # Grab the next corresponding pair of snapshots.
            x_t = X[:, ind_t][..., None]
            y_t = Y[:, ind_t][..., None]

            # Get the results of this Cholesky factorization iteration.
            if self._online:
                cholesky_results = self._cholesky_step(
                    x_t, kernel_function, self._cholesky
                )
            else:
                cholesky_results = self._cholesky_step(x_t, kernel_function, C)
            _, s_t, _, k_tt, delta_t = cholesky_results

            # NOT almost linearly dependent - add x to the dictionary.
            if np.abs(delta_t) > self._dict_tol:
                self._sparse_dictionary = np.hstack(
                    [self._sparse_dictionary, x_t]
                )

                # Update the Cholesky factor.
                if self._online:
                    self._cholesky = self._update_cholesky(
                        self._cholesky, s_t, k_tt
                    )
                    self._update_online(
                        y_t, cholesky_results, cholesky_updated=True
                    )
                else:
                    C = self._update_cholesky(C, s_t, k_tt)

                if k_tt < np.sum(s_t**2):
                    msg = (
                        "The Cholesky factor is ill-conditioned. Consider "
                        "increasing dict_tol or changing the kernel function."
                    )
                    warnings.warn(msg)

            # Online learning updates for the almost linearly dependent case.
            elif self._online:
                self._update_online(
                    y_t, cholesky_results, cholesky_updated=False
                )

        # Compute weights in one go, if not performing online learning.
        if not self._online:
            K_mat = kernel_function(self._sparse_dictionary, X)
            if self._lstsq:  # use least squares
                self._weights = np.linalg.lstsq(K_mat.T, Y.T, rcond=None)[0].T
            else:  # use the pseudo-inverse
                self._weights = Y.dot(np.linalg.pinv(K_mat))

    def compute_linear_operator(
        self, fixed_point, compute_A, kernel_gradient, x_rescale
    ):
        """
        Compute the DMD diagnostics of the linear model at a fixed point.

        :param fixed_point: base state about which to perturb the system.
        :type fixed_point: numpy.ndarray
        :param compute_A: if True, the full linear operator is explicitly
            computed and stored. If False, the full linear operator isn't
            computed and stored explicitly.
        :type compute_A: bool
        :param kernel_gradient: gradient of the kernel function applied.
        :type kernel_gradient: function
        :param x_rescale: value or (n_features,) array of values for
            rescaling the features of the input data.
        :type x_rescale: float or numpy.ndarray
        """
        kernel_grad = kernel_gradient(self._sparse_dictionary, fixed_point)
        kernel_grad = np.multiply(kernel_grad, x_rescale)
        U, s, V = compute_svd(kernel_grad.T, self._svd_rank)
        if compute_A:
            self._A = self.weights.dot(kernel_grad)
            self._Atilde = np.linalg.multi_dot([U.conj().T, self._A, U])
        else:
            self._Atilde = np.linalg.multi_dot(
                [U.conj().T, self.weights, kernel_grad, U]
            )

        # Filter out zero eigenvalues.
        eigenvalues, eigenvectors = np.linalg.eig(self._Atilde)
        nonzero_inds = np.abs(eigenvalues) > 1e-16
        eigenvalues = eigenvalues[nonzero_inds]
        eigenvectors = eigenvectors[:, nonzero_inds]

        # Sort eigenvalues, descending based on modulus.
        sorted_inds = np.argsort(-np.abs(eigenvalues))
        self._eigenvalues = eigenvalues[sorted_inds]
        self._eigenvectors = eigenvectors[:, sorted_inds]

        self._modes = np.linalg.multi_dot(
            [
                self._weights,
                V,
                np.diag(s),
                self._eigenvectors,
                np.diag(1 / self._eigenvalues),
            ]
        )

    @staticmethod
    def _update_cholesky(cholesky, s, kxx):
        """
        Helper function that updates the cholesky factor given the current
        cholesky factor and the necessary quantities for updating. See the
        documentation of `_cholesky_step()` for more parameter information.
        """
        cholesky = np.vstack(
            [
                np.hstack([cholesky, np.zeros((len(cholesky), 1))]),
                np.append(
                    s.conj().T,
                    max(0, np.abs(np.sqrt(kxx - np.sum(s**2)))),
                ),
            ]
        )
        return cholesky

    def _cholesky_step(self, x, kernel_function, cholesky):
        """
        Helper function that computes and returns all of the quantities used to
        decide the progression of the Cholesky factorization routine. See the
        original reference for specific equations and notations.
        """
        # Equation (3.11): Evaluate the kernel using the current dictionary
        # items and the next candidate addition to the dictionary.
        k_tilde = kernel_function(self._sparse_dictionary, x)

        # Equation (3.10): Use backsubstitution to compute the span of the
        # current dictionary.
        s = np.linalg.lstsq(cholesky, k_tilde, rcond=None)[0]
        pi = np.linalg.lstsq(cholesky.conj().T, s, rcond=None)[0]

        # Equation (3.9): Compute the minimum (squared) distance between
        # the current sample and the span of the current dictionary.
        kxx = kernel_function(x, x)[0][0]
        delta = kxx - k_tilde.conj().T.dot(pi)

        return k_tilde, s, pi, kxx, delta

    def _update_online(self, y, cholesky_results, cholesky_updated):
        """
        Helper function that performs a single online learning update for
        both the almost linearly dependent case and the not almost linearly
        dependent case. Updates the P matrix and the weight matrix. See the
        supplemental information of the original reference for more details.
        """
        k_tilde, _, pi, _, delta = cholesky_results

        # NOT almost linearly dependent case.
        if cholesky_updated:
            self._P = np.vstack(
                [
                    np.hstack([self._P, np.zeros((len(self._P), 1))]),
                    np.append(np.zeros((1, len(self._P))), 1.0),
                ]
            )
            weights_update = (y - self._weights.dot(k_tilde)) / delta
            self._weights = np.hstack(
                [
                    self._weights - weights_update.dot(pi.conj().T),
                    weights_update,
                ]
            )
        # Almost linearly dependent case.
        else:
            h = pi.conj().T.dot(self._P) / (
                1.0 + np.linalg.multi_dot([pi.conj().T, self._P, pi])
            )
            self._P = self._P.dot(np.eye(len(self._P)) - pi.dot(h))
            weights_update = np.linalg.lstsq(
                (self._cholesky.dot(self._cholesky.conj().T)).T,
                ((y - self._weights.dot(k_tilde)).dot(h)).T,
                rcond=None,
            )[0].T
            self._weights += weights_update


class LANDO(DMDBase):
    """
    Linear and nonlinear disambiguation optimization (LANDO).

    :param svd_rank: the rank for the truncation of the linear operator; If 0,
        the method computes the optimal rank and uses it for truncation; if a
        positive integer, the method uses the argument for the truncation; if
        float between 0 and 1, the rank is the number of the biggest singular
        values that are needed to reach the 'energy' specified by `svd_rank`;
        if -1, the method does not compute a truncation.
    :type svd_rank: int or float
    :param tlsq_rank: rank truncation computing Total Least Square. Default is
        0, which means no truncation.
    :type tlsq_rank: int
    :param opt: If True, amplitudes are computed like in optimized DMD  (see
        :func:`~dmdbase.DMDBase._compute_amplitudes` for reference). If
        False, amplitudes are computed following the standard algorithm. If
        `opt` is an integer, it is used as the (temporal) index of the snapshot
        used to compute DMD modes amplitudes (following the standard
        algorithm).
        The reconstruction will generally be better in time instants near the
        chosen snapshot; however increasing `opt` may lead to wrong results
        when the system presents small eigenvalues. For this reason a manual
        selection of the number of eigenvalues considered for the analyisis may
        be needed (check `svd_rank`). Also setting `svd_rank` to a value
        between 0 and 1 may give better results. Default is False.
    :type opt: bool or int
    :param kernel_metric: the kernel function to apply. Supported kernel
        metrics for `LANDO` are `"linear"`, `"poly"`, and `"rbf"`.
    :type kernel_metric: str
    :param kernel_params: additional parameters to be inputted to the
        `sklearn.metrics.pairwise_kernels` function. This includes
        kernel-specific function parameters.
    :type kernel_params: dict
    :param kernel_function: custom kernel function to be used instead of the
        `sklearn.metrics.pairwise_kernels` built-ins. Must be able to take a
        (n_features, n_samples_X) numpy.ndarray and a (n_features, n_samples_Y)
        numpy.ndarray and return the (n_samples_X, n_samples_Y) kernel matrix
        as a numpy.ndarray. In other words, for the input matrices X, Y,
            kernel_function(X, Y) = K(X, Y),
        where K(X, Y) denotes the kernel matrix.
    :type kernel_function: function
    :param kernel_gradient: gradient of the kernel function evaluated at the
        given (n_features, n_samples_X) numpy.ndarray, and at the given
        (n_features,) numpy.ndarray. Must return a (n_samples_X, n_features)
        numpy.ndarray. This function must be defined in order to perform
        `analyze_fixed_point()` if a custom kernel function is given.
        In other words, for the input matrix X and input vector y,
            kernel_gradient(X, y) = \nabla K(X, y)
        for the kernel function defined via the kernel_function input.
    :type kernel_gradient: function
    :param x_rescale: value or `snapshots_shape` numpy array of values for
        rescaling the features of the input data. Can be used to improve
        conditioning.
    :type x_rescale: float or numpy.ndarray
    :param dict_tol: threshold at which delta_t, the degree to which a snapshot
        x_t can be represented by the snapshot dictionary in feature space, is
        considered high enough that x_t should be added to the snapshot
        dictionary. Referred to as \nu in the original paper. Note that
        increasing this tolerance will lead to sparser dictionaries.
    :type dict_tol: float
    :param online: whether or not to use the online learning variant of LANDO.
        Default is False, to not use the online learning algorithm.
    :type online: bool
    :param permute: whether or not to randomly permute the order of the
        snapshots prior to learning the snapshot dictionary. Default is True,
        to permute the snapshots, as doing so is generally recommended in order
        to avoid generating ill-conditioned dictionaries.
    :type permute: bool
    :param lstsq: method used for computing the weights of the dictionary-based
        kernel model. If True, least-squares is used to solve for the weights,
        otherwise the pseudo-inverse is used. Note that this parameter is
        ignored if online learning is enabled.
    :type lstsq: bool
    """

    def __init__(
        self,
        svd_rank=0,
        tlsq_rank=0,
        opt=False,
        kernel_metric="linear",
        kernel_params=None,
        kernel_function=None,
        kernel_gradient=None,
        x_rescale=1.0,
        dict_tol=1e-6,
        online=False,
        permute=True,
        lstsq=True,
    ):
        # Store the supported kernel metrics.
        self._supported_kernels = ("linear", "poly", "rbf")

        # Check the validity of the provided kernel functions.
        if kernel_params is None:
            kernel_params = {}
        self._test_kernel_inputs(kernel_metric, kernel_params)
        self._test_kernel_functions(kernel_function, kernel_gradient)

        # Initialize the basic DMD attributes.
        super().__init__(
            svd_rank=svd_rank,
            tlsq_rank=tlsq_rank,
            exact=True,
            opt=opt,
        )

        # Set the kernel functions, which must be valid.
        self._kernel_metric = kernel_metric
        self._kernel_params = kernel_params
        self._kernel_function = kernel_function
        self._kernel_gradient = kernel_gradient

        # Build the LANDO operator.
        self._Atilde = LANDOOperator(
            svd_rank=svd_rank,
            dict_tol=dict_tol,
            online=online,
            permute=permute,
            lstsq=lstsq,
        )

        # Keep track of the data rescaling factor.
        self._x_rescale = x_rescale

        # Keep track of the last fixed point analysis.
        self._compute_A = False
        self._fixed_point = None
        self._bias = None

        # Keep track of whether or not the model was ever updated.
        self._updated = False

    def kernel_function(self, X, Y):
        """
        Calls `sklearn.metrics.pairwise_kernels` to evaluate the kernel
        function at the given feature matrices X and Y. If a custom kernel
        function was provided, then that function will be used instead.

        :param X: feature matrix with shape (n_features, n_samples_X)
        :type X: numpy.ndarray
        :param Y: second feature matrix with shape (n_features, n_samples_Y)
        :type Y: numpy.ndarray
        :return: kernel matrix with shape (n_samples_X, n_samples_Y)
        :rtype: numpy.ndarray
        """
        if self._kernel_function is not None:
            return self._kernel_function(X, Y)
        return pairwise_kernels(
            X.T, Y.T, metric=self._kernel_metric, **self._kernel_params
        )

    def kernel_gradient(self, X, y):
        """
        Computes the gradient of the kernel with respect to the second feature
        vector x. Then evaluates the gradient at the feature matrix X and the
        feature vector y. Currently, this method is only compatible with
        the linear, polynomial, and RBF kernels, unless a custom kernel
        gradient function was provided.

        :param X: feature matrix with shape (n_features, n_samples_X)
        :type X: numpy.ndarray
        :param y: feature vector with shape (n_features,)
        :type y: numpy.ndarray
        :return: kernel gradient matrix with shape (n_samples_X, n_features)
        :rtype: numpy.ndarray
        """
        if self._kernel_function is not None:
            if self._kernel_gradient is None:
                msg = (
                    "Unable to evaluate the gradient of the kernel function. "
                    "If a custom kernel function was provided, please ensure "
                    "that the corresponding gradient function is also defined "
                    "by setting the kernel_gradient parameter the LANDO model."
                )
                raise ValueError(msg)
            return self._kernel_gradient(X, y)

        if self._kernel_metric == "linear":
            return X.T

        # Kernel metric must be polynomial or RBF.
        if "gamma" in self._kernel_params:
            gamma = self._kernel_params["gamma"]
        else:  # set the pairwise_kernels gamma default
            gamma = 1.0 / X.shape[0]

        if self._kernel_metric == "poly":
            if "coef0" in self._kernel_params:
                coef0 = self._kernel_params["coef0"]
            else:
                coef0 = 1
            if "degree" in self._kernel_params:
                degree = self._kernel_params["degree"]
            else:
                degree = 3
            return np.diag(
                gamma * degree * (coef0 + gamma * X.T.dot(y)) ** (degree - 1)
            ).dot(X.T)

        # Kernel metric is RBF.
        centered_X = X - y[..., None]
        return np.diag(
            -2
            * gamma
            * np.exp(-gamma * np.linalg.norm(centered_X, 2, axis=0) ** 2)
        ).dot(centered_X.T)

    @property
    def supported_kernels(self):
        """
        Get the `sklearn.metrics.pairwise.pairwise_kernels()` metrics
        that are currently compatible with the `LANDO` module.

        :return: currently available kernel metrics.
        :rtype: tuple(str,str,str)
        """
        return self._supported_kernels

    @property
    def partially_fitted(self):
        """
        Check whether the weights for this LANDO instance have been computed.
        Note that a LANDO model is only fully fitted once the weights have been
        computed, AND an analysis about a fixed point has been performed.

        :return: `True` if instance has been partially fitted, else `False`.
        :rtype: bool
        """
        try:
            return self.operator.weights is not None
        except ValueError:
            return False

    @property
    def sparse_dictionary(self):
        """
        Get the learned sparse feature dictionary.

        :return: the sparse feature dictionary.
        :rtype: numpy.ndarray
        """
        # Note: the attribute operator.sparse_dictionary is re-scaled.
        # Thus we undo this, but only when the user asks for the dictionary.
        if not self.partially_fitted:
            raise ValueError("You need to call fit().")
        if isinstance(self._x_rescale, np.ndarray):
            return np.divide(
                self.operator.sparse_dictionary,
                self._x_rescale[..., None],
            )
        return self.operator.sparse_dictionary / self._x_rescale

    def f(self, X):
        """
        Prediction of the true system model F, where F(x) = y.

        :param X: feature vectors from the data domain.
        :type X: numpy.ndarray or iterable
        :return: f(X), an approximation of the true quantity F(X) = Y.
        :rtype: numpy.ndarray
        """
        if not self.partially_fitted:
            raise ValueError("You need to call fit().")

        # Ensure that the input is an array whose columns contain snapshots
        # with a dimension that matches that of the original input data.
        X = np.array(X)
        if X.shape[:-1] != self.snapshots_shape:
            msg = "Last dimension of X must contain data with shape {}."
            raise ValueError(msg.format(self.snapshots_shape))

        # Flatten and rescale the snapshots before applying the model.
        X = X.reshape(-1, X.shape[-1])
        X = self._rescale(X)

        return self.operator.weights.dot(
            self.kernel_function(self.operator.sparse_dictionary, X)
        )

    @property
    def fixed_point(self):
        """
        Get the fixed point from the last fixed point analysis.

        :return: the fixed point from the last fixed point analysis.
        :rtype: numpy.ndarray
        """
        if not self.fitted:
            msg = "You need to call fit() and analyze_fixed_point()."
            raise ValueError(msg)
        return self._fixed_point

    @property
    def bias(self):
        """
        Get the bias term from the last fixed point analysis. This is in
        reference to the term c in the expression
            f(x) = c + Ax' + N(x'), x = x_bar + x'
        for our system perturbed about x_bar. f is our model approximation,
        A is a linear operator, and N is a nonlinear operator.

        :return: the bias term from the last fixed point analysis.
        :rtype: numpy.ndarray
        """
        if not self.fitted:
            msg = "You need to call fit() and analyze_fixed_point()."
            raise ValueError(msg)
        return self._bias

    @property
    def linear(self):
        """
        Get the linear operator from the last fixed point analysis. This is in
        reference to the term A in the expression
            f(x) = c + Ax' + N(x'), x = x_bar + x'
        for our system perturbed about x_bar. f is our model approximation,
        c is a constant bias term, and N is a nonlinear operator.

        :return: the linear operator from the last fixed point analysis.
        :rtype: numpy.ndarray
        """
        if not self.fitted:
            msg = "You need to call fit() and analyze_fixed_point()."
            raise ValueError(msg)
        return self.operator.A

    def nonlinear(self, X):
        """
        Get the nonlinear term from the last fixed point analysis, evaluated
        at the input snapshots X. This is in reference to the term N(x') in
        the expression
            f(x) = c + Ax' + N(x'), x = x_bar + x'
        for our system perturbed about x_bar. f is our model approximation,
        c is a constant bias term, and A is a linear operator.

        :param X: feature vectors from the data domain.
        :type X: numpy.ndarray or iterable
        :return: the nonlinear terms from the last fixed point analysis,
            evaluated at the input vectors.
        :rtype: numpy.ndarray
        """
        if not self.fitted:
            msg = "You need to call fit() and analyze_fixed_point()."
            raise ValueError(msg)

        # Ensure that the input is an array whose columns contain snapshots
        # with a dimension that matches that of the original input data.
        X = np.array(X)
        if X.shape[:-1] != self.snapshots_shape:
            msg = "Last dimension of X must contain data with shape {}."
            raise ValueError(msg.format(self.snapshots_shape))

        # Re-create the kernel gradient matrix for the linear operator.
        kernel_grad = np.multiply(
            self.kernel_gradient(
                self.operator.sparse_dictionary,
                self._rescale(self._fixed_point.flatten()),
            ),
            self._x_rescale,
        )

        # Define the f(x) term.
        fX = self.f(X + self._fixed_point[..., None])

        return (
            fX
            - self._bias
            - np.linalg.multi_dot(
                [
                    self.operator.weights,
                    kernel_grad,
                    X.reshape(-1, X.shape[-1]),
                ]
            )
        )

    def fit(self, X, Y=None):
        """
        Compute the linear and nonlinear disambiguation optimization (LANDO).

        :param X: the input snapshots.
        :type X: numpy.ndarray or iterable
        :param Y: additional input snapshots such that F(x) = y. If not given,
            snapshots from X are used to build a discrete-time model.
        :type Y: numpy.ndarray or iterable
        """
        self._reset()
        self._snapshots_holder = Snapshots(X)
        self._check_x_rescale()

        if Y is None:
            X = self.snapshots[:, :-1]
            Y = self.snapshots[:, 1:]
        else:
            self._compare_data_shapes(Snapshots(Y).snapshots)
            self._snapshots_holder_y = Snapshots(Y)
            X = self.snapshots
            Y = self.snapshots_y

        X, Y = compute_tlsq(X, Y, self._tlsq_rank)
        X_rescaled = self._rescale(X)
        self.operator.compute_operator(X_rescaled, Y, self.kernel_function)

        # Default timesteps
        self._set_initial_time_dictionary(
            {"t0": 0, "tend": self.snapshots.shape[1] - 1, "dt": 1}
        )

        return self

    def analyze_fixed_point(self, fixed_point, compute_A=False):
        """
        Use a partially-fitted LANDO model to examine the system perturbed
        about a given base state x_bar such that x = x_bar + x'. Extract the
        bias term, linear portion, and nonlinear portion from the expression
            f(x) = c + Ax' + N(x'),
        where f is our model approximation, c is a constant bias term, A is a
        linear operator, and N is a nonlinear operator. Additionally obtain the
        DMD diagnostics of the linear model A, along with the DMD amplitudes
        associated with the computed linear model.

        :param fixed_point: base state about which to analyze the system.
        :type fixed_point: numpy.ndarray or iterable
        :param compute_A: whether or not to explicitly compute the full linear
            operator A. If True, A is computed and stored explicitly, otherwise
            A is not stored and is only computed when needed.
        :type compute_A: bool
        """
        if not self.partially_fitted:
            msg = (
                "You need to call fit() before "
                "performing a fixed point analysis."
            )
            raise ValueError(msg)

        fixed_point = np.array(fixed_point)
        if fixed_point.shape != self.snapshots_shape:
            msg = "Input fixed point must have shape {}."
            raise ValueError(msg.format(self.snapshots_shape))

        self._compute_A = compute_A
        self._fixed_point = fixed_point
        self._bias = self.f(fixed_point[..., None])
        self.operator.compute_linear_operator(
            self._rescale(fixed_point.flatten()),
            self._compute_A,
            self.kernel_gradient,
            self._x_rescale,
        )

        self._b = self._compute_amplitudes()

    def update(self, X, Y=None):
        """
        Update a fitted LANDO model using new data.

        Note that this function only updates the LANDO model, and not the
        stored snapshots or DMDTimeDicts. Hence be aware that using this
        function may result in strange or unexpected behavior when used
        in conjunction with certain DMDBase functionalities.

        :param X: the input snapshots.
        :type X: numpy.ndarray or iterable
        :param Y: additional input snapshots such that F(x) = y. If not given,
            snapshots from X are used to build a discrete-time model.
        :type Y: numpy.ndarray or iterable
        """
        if not self.partially_fitted:
            msg = "You need to call fit() before updating a LANDO model."
            raise ValueError(msg)

        X_newsnap = Snapshots(X).snapshots
        if Y is None:
            X = X_newsnap[:, :-1]
            Y = X_newsnap[:, 1:]
        else:
            Y_newsnap = Snapshots(Y).snapshots
            if X_newsnap.shape != Y_newsnap.shape:
                msg = "X {} and Y {} input data must be the same shape."
                raise ValueError(msg.format(X_newsnap.shape, Y_newsnap.shape))
            X = X_newsnap
            Y = Y_newsnap

        X, Y = compute_tlsq(X, Y, self._tlsq_rank)
        X_rescaled = self._rescale(X)
        self.operator.compute_operator(
            X_rescaled, Y, self.kernel_function, updating=True
        )

        # If a fixed point analysis was already done, redo it.
        if self.fitted:
            self._bias = self.f(self._fixed_point[..., None])
            self.operator.compute_linear_operator(
                self._rescale(self._fixed_point.flatten()),
                self._compute_A,
                self.kernel_gradient,
                self._x_rescale,
            )
            self._b = self._compute_amplitudes()

        # Flag this model as updated.
        self._updated = True

        return self

    def predict(self, x0, tend, continuous=True, dt=1.0, solve_ivp_opts=None):
        """
        Reconstruct or predict the state of the system using the fitted model.

        :param x0: initial condition from which to propagate.
        :type x0: numpy.ndarray or iterable
        :param tend: number of data points to compute.
        :type tend: int
        :param continuous: if True, the method assumes that the fitted model
            f(x) is continuous, in which case f(x) predicts the time derivative
            x'. If False, the method assumes that the fitted model is discrete,
            in which case f(x_n) predicts the the future data point x_{n+1}.
        :type continuous: bool
        :param dt: desired time step between each computed data point. This
            parameter is only used if `continuous=True`.
        :type dt: float
        :param solve_ivp_opts: dictionary of additional parameters to be passed
            to the `scipy.integrate.solve_ivp` function. This parameter is only
            used if `continuous=True`.
        :type solve_ivp_opts: dict
        """
        if not self.partially_fitted:
            raise ValueError("You need to call fit().")

        x0 = np.array(x0)
        if x0.shape != self.snapshots_shape:
            msg = "Input initial condition must have shape {}."
            raise ValueError(msg.format(self.snapshots_shape))
        x0 = x0.flatten()

        if continuous:

            def ode_sys(xt, x):
                return self.f(x[..., None]).flatten()

            if solve_ivp_opts is None:
                solve_ivp_opts = {}

            t_eval = np.arange(0, tend * dt, dt)

            sol = solve_ivp(
                ode_sys,
                [t_eval[0], t_eval[-1]],
                x0,
                t_eval=t_eval,
                **solve_ivp_opts,
            )

            return sol.y

        # Otherwise, assume a discrete-time system.
        Y = np.empty((len(x0), tend))
        Y[:, 0] = x0
        for i in range(tend - 1):
            Y[:, i + 1] = self.f(Y[:, i][..., None]).flatten()

        return Y

    @property
    def reconstructed_data(self):
        """
        Get the DMD reconstruction of the data.

        :return: the DMD reconstruction.
        :rtype: numpy.ndarray
        """
        msg = "Consider using the predict() method for system reconstructions."
        warnings.warn(msg)
        return self.modes.dot(self.dynamics)

    def _rescale(self, X):
        """
        Helper function that rescales the given data according to
        the current `LANDO` instance's `x_rescale` value(s).
        """
        if isinstance(self._x_rescale, np.ndarray):
            if np.ndim(X) == 1:
                return np.multiply(X, self._x_rescale)
            return np.multiply(X, self._x_rescale[..., None])
        return X * self._x_rescale

    def _check_x_rescale(self):
        """
        Helper function that ensures that `x_rescale` is either an int, float,
        or a numpy array. Also ensures that if `x_rescale` is an array, then it
        must possess the same shape as the flattened input snapshots.
        """
        if not isinstance(self._x_rescale, (int, float, np.ndarray)):
            raise TypeError("x_rescale must be a float or a numpy array.")
        if isinstance(self._x_rescale, np.ndarray):
            if self._x_rescale.shape != self.snapshots_shape:
                msg = (
                    "If a numpy array, x_rescale must have the "
                    "same shape {} as the input features X."
                )
                raise ValueError(msg.format(self.snapshots_shape))
            self._x_rescale = self._x_rescale.flatten()

    def _test_kernel_inputs(self, kernel_metric, kernel_params):
        """
        Helper function that checks the validity of the provided kernel metric
        and parameters. Additionally uses a dummy array of data in order to
        call `sklearn.metrics.pairwise_kernels` using the given parameters.
        """
        if kernel_metric not in self._supported_kernels:
            msg = (
                "Invalid kernel metric '{}' provided. "
                "Please use one of the following metrics: {}"
            )
            raise ValueError(msg.format(kernel_metric, self._supported_kernels))

        if not isinstance(kernel_params, dict):
            raise TypeError("kernel_params must be a dict.")

        # Test that sklearn.metrics.pairwise_kernels can be called.
        x_dummy = np.arange(4).reshape(2, 2)
        try:
            pairwise_kernels(
                x_dummy, x_dummy, metric=kernel_metric, **kernel_params
            )
        except Exception as e:
            msg = (
                "Invalid kernel parameters {} provided. "
                "Please verify that kernel_params are passable to "
                "sklearn.metrics.pairwise_kernels with metric '{}'."
            )
            raise ValueError(msg.format(kernel_params, kernel_metric)) from e

    @staticmethod
    def _test_kernel_functions(kernel_function, kernel_gradient):
        """
        Helper function that checks the validity of the provided kernel
        functions. Ensures that either both are None, or both inputs are valid.
        In short...
        - If provided, kernel_function and kernel_gradient must be functions
            that accept and return numpy arrays of the expected shape.
        - kernel_gradient function cannot be provided without kernel_function.
        - kernel_function can be provided without a kernel_gradient function,
            however a fixed point analysis cannot be performed as a result.
        """
        # Test moves on if both inputs are None.
        if kernel_function is not None or kernel_gradient is not None:
            # kernel gradient can't be provided alone.
            if kernel_function is None:
                msg = (
                    "kernel_gradient cannot be provided "
                    "without a corresponding kernel_function."
                )
                raise ValueError(msg)
            # kernel_function needs to be a function.
            if not isfunction(kernel_function):
                raise TypeError("kernel_function must be a function.")
            # Warn the user if kernel_gradient wasn't provided.
            if kernel_gradient is None:
                msg = (
                    "kernel_function given without kernel_gradient. Note that "
                    "fit() may be performed, but kernel_gradient must now be "
                    "provided in order to perform analyze_fixed_point()."
                )
                warnings.warn(msg)
            # kernel_gradient needs to be a function.
            elif not isfunction(kernel_gradient):
                raise TypeError("kernel_gradient must be a function")

            # Test that the given functions take and yield what is expected.
            X_dummy = np.empty((5, 10))
            Y_dummy = np.empty((5, 20))
            general_msg = (
                "Please check the LANDO class documentation "
                "for details on how to format the functions "
                "kernel_function and kernel_gradient."
            )
            # Test the kernel function.
            try:
                K_xy = kernel_function(X_dummy, Y_dummy)
            except Exception as e:
                msg = "Error calling kernel_function. "
                raise ValueError(msg + general_msg) from e
            if not isinstance(K_xy, np.ndarray) or K_xy.shape != (10, 20):
                msg = "kernel_function must return a {} numpy array. "
                raise ValueError((msg + general_msg).format((10, 20)))

            # Test the kernel gradient function.
            if kernel_gradient is not None:
                try:
                    grad = kernel_gradient(X_dummy, Y_dummy[:, 0])
                except Exception as e:
                    msg = "Error calling kernel_gradient. "
                    raise ValueError(msg + general_msg) from e
                if not isinstance(grad, np.ndarray) or grad.shape != (10, 5):
                    msg = "kernel_gradient must return a {} numpy array. "
                    raise ValueError((msg + general_msg).format((10, 5)))



"""
Derived module from dmdbase.py for dmd with control.

Reference:
- Proctor, J.L., Brunton, S.L. and Kutz, J.N., 2016. Dynamic mode decomposition
with control. SIAM Journal on Applied Dynamical Systems, 15(1), pp.142-161.
"""

import numpy as np

from .dmdbase import DMDBase
from .dmdoperator import DMDOperator
from .snapshots import Snapshots
from .utils import compute_svd, compute_tlsq


class DMDControlOperator(DMDOperator):
    """
    DMD with control base operator. This should be subclassed in order to
    implement the appropriate features.

    :param svd_rank: the rank for the truncation; If 0, the method computes the
        optimal rank and uses it for truncation; if positive interger, the
        method uses the argument for the truncation; if float between 0 and 1,
        the rank is the number of the biggest singular values that are needed
        to reach the 'energy' specified by `svd_rank`; if -1, the method does
        not compute truncation.
    :type svd_rank: int or float
    :param svd_rank_omega: the rank for the truncation of the aumented matrix
        omega composed by the left snapshots matrix and the control. Used only
        for the `_fit_B_unknown` method of this class. It should be greater or
        equal than `svd_rank`. For the possible values please refer to the
        `svd_rank` parameter description above.
    :type svd_rank_omega: int or float
    :param int tlsq_rank: rank truncation computing Total Least Square. Default
        is 0, that means no truncation.
    """

    def __init__(self, svd_rank, svd_rank_omega, tlsq_rank):
        super(DMDControlOperator, self).__init__(
            svd_rank=svd_rank,
            exact=True,
            rescale_mode=None,
            forward_backward=False,
            sorted_eigs=False,
            tikhonov_regularization=None,
        )
        self._svd_rank_omega = svd_rank_omega
        self._tlsq_rank = tlsq_rank


class DMDBKnownOperator(DMDControlOperator):
    """
    DMD with control base operator when B is given.

    :param svd_rank: the rank for the truncation; If 0, the method computes the
        optimal rank and uses it for truncation; if positive interger, the
        method uses the argument for the truncation; if float between 0 and 1,
        the rank is the number of the biggest singular values that are needed
        to reach the 'energy' specified by `svd_rank`; if -1, the method does
        not compute truncation.
    :type svd_rank: int or float
    :param svd_rank_omega: the rank for the truncation of the aumented matrix
        omega composed by the left snapshots matrix and the control. Used only
        for the `_fit_B_unknown` method of this class. It should be greater or
        equal than `svd_rank`. For the possible values please refer to the
        `svd_rank` parameter description above.
    :type svd_rank_omega: int or float
    :param int tlsq_rank: rank truncation computing Total Least Square. Default
        is 0, that means no truncation.
    """

    def compute_operator(self, X, Y, B, controlin):
        """
        Compute the low-rank operator. This is the standard version of the DMD
        operator, with a correction which depends on B.

        :param numpy.ndarray X: matrix containing the snapshots x0,..x{n-1} by
            column.
        :param numpy.ndarray Y: matrix containing the snapshots x1,..x{n} by
            column.
        :param numpy.ndarray B: the matrix B.
        :param numpy.ndarray control: the control input.
        :return: the (truncated) left-singular vectors matrix, the (truncated)
            singular values array, the (truncated) right-singular vectors
            matrix of X.
        :rtype: numpy.ndarray, numpy.ndarray, numpy.ndarray
        """
        X, Y = compute_tlsq(X, Y, self._tlsq_rank)
        Y = Y - B.dot(controlin)
        return super(DMDBKnownOperator, self).compute_operator(X, Y)


class DMDBUnknownOperator(DMDControlOperator):
    """
    DMD with control base operator when B is unknown.

    :param svd_rank: the rank for the truncation; If 0, the method computes the
        optimal rank and uses it for truncation; if positive interger, the
        method uses the argument for the truncation; if float between 0 and 1,
        the rank is the number of the biggest singular values that are needed
        to reach the 'energy' specified by `svd_rank`; if -1, the method does
        not compute truncation.
    :type svd_rank: int or float
    :param svd_rank_omega: the rank for the truncation of the aumented matrix
        omega composed by the left snapshots matrix and the control. Used only
        for the `_fit_B_unknown` method of this class. It should be greater or
        equal than `svd_rank`. For the possible values please refer to the
        `svd_rank` parameter description above.
    :type svd_rank_omega: int or float
    :param int tlsq_rank: rank truncation computing Total Least Square. Default
        is 0, that means no truncation.
    """

    def compute_operator(self, X, Y, controlin):
        """
        Compute the low-rank operator.

        :param numpy.ndarray X: matrix containing the snapshots x0,..x{n-1} by
            column.
        :param numpy.ndarray Y: matrix containing the snapshots x1,..x{n} by
            column.
        :param numpy.ndarray control: the control input.
        :return: the (truncated) left-singular vectors matrix of Y, and
            the product between the left-singular vectors of Y and Btilde.
        :rtype: numpy.ndarray, numpy.ndarray
        """
        snapshots_rows = X.shape[0]

        omega = np.vstack([X, controlin])

        Up, sp, Vp = compute_svd(omega, self._svd_rank_omega)

        Up1 = Up[:snapshots_rows, :]
        Up2 = Up[snapshots_rows:, :]

        Ur, _, _ = compute_svd(Y, self._svd_rank)

        self._Atilde = np.linalg.multi_dot(
            [Ur.T.conj(), Y, Vp, np.diag(np.reciprocal(sp)), Up1.T.conj(), Ur]
        )
        self._compute_eigenquantities()
        self._compute_modes(Y, sp, Vp, Up1, Ur)

        Btilde = np.linalg.multi_dot(
            [Ur.T.conj(), Y, Vp, np.diag(np.reciprocal(sp)), Up2.T.conj()]
        )

        return Ur, Ur.dot(Btilde)

    def _compute_modes(self, Y, sp, Vp, Up1, Ur):
        """
        Private method that computes eigenvalues and eigenvectors of the
        high-dimensional operator (stored in self.modes and self.Lambda).
        """

        self._modes = np.linalg.multi_dot(
            [
                Y,
                Vp,
                np.diag(np.reciprocal(sp)),
                Up1.T.conj(),
                Ur,
                self.eigenvectors,
            ]
        )
        self._Lambda = self.eigenvalues


class DMDc(DMDBase):
    """
    Dynamic Mode Decomposition with control.
    This version does not allow to manipulate the temporal window within the
    system is reconstructed.

    :param svd_rank: the rank for the truncation; If 0, the method computes the
        optimal rank and uses it for truncation; if positive interger, the
        method uses the argument for the truncation; if float between 0 and 1,
        the rank is the number of the biggest singular values that are needed
        to reach the 'energy' specified by `svd_rank`; if -1, the method does
        not compute truncation.
    :type svd_rank: int or float
    :param int tlsq_rank: rank truncation computing Total Least Square. Default
        is 0, that means no truncation.
    :param opt: argument to control the computation of DMD modes amplitudes.
        See :class:`DMDBase`. Default is False.
    :type opt: bool or int
    :param svd_rank_omega: the rank for the truncation of the aumented matrix
        omega composed by the left snapshots matrix and the control. Used only
        for the `_fit_B_unknown` method of this class. It should be greater or
        equal than `svd_rank`. For the possible values please refer to the
        `svd_rank` parameter description above.
    :type svd_rank_omega: int or float
    :param lag: the time lag for the snapshots. Used in fit method to generate
        X and Y.
    :type lag: int
    """

    def __init__(
        self, svd_rank=0, tlsq_rank=0, opt=False, svd_rank_omega=-1, lag=1
    ):
        # we're going to initialize Atilde when we know if B is known
        self._Atilde = None
        # remember the arguments for when we'll need them
        self._dmd_operator_kwargs = {
            "svd_rank": svd_rank,
            "svd_rank_omega": svd_rank_omega,
            "tlsq_rank": tlsq_rank,
        }

        self._opt = opt
        self._exact = False

        self._B = None
        self._snapshots_holder = None
        self._controlin = None
        self._basis = None
        self._lag = lag

        self._modes_activation_bitmask_proxy = None

    @property
    def svd_rank_omega(self):
        return self.operator._svd_rank_omega

    @property
    def B(self):
        """
        Get the operator B.

        :return: the operator B.
        :rtype: numpy.ndarray
        """
        return self._B

    @property
    def basis(self):
        """
        Get the basis used to reduce the linear operator to the low dimensional
        space.

        :return: the matrix which columns are the basis vectors.
        :rtype: numpy.ndarray
        """
        return self._basis

    def reconstructed_data(self, control_input=None):
        """
        Return the reconstructed data, computed using the `control_input`
        argument. If the `control_input` is not passed, the original input (in
        the `fit` method) is used. The input dimension has to be consistent
        with the dynamics.

        :param numpy.ndarray control_input: the input control matrix.
        :return: the matrix that contains the reconstructed snapshots.
        :rtype: numpy.ndarray
        """
        controlin = (
            np.asarray(control_input)
            if control_input is not None
            else self._controlin
        )

        if controlin.shape[-1] != self.dynamics.shape[-1] - self._lag:
            raise RuntimeError(
                "The number of control inputs and the number of snapshots to "
                "reconstruct has to be the same"
            )

        eigs = np.power(
            self.eigs, self.dmd_time["dt"] // self.original_time["dt"]
        )
        A = np.linalg.multi_dot(
            [self.modes, np.diag(eigs), np.linalg.pinv(self.modes)]
        )

        data = [self.snapshots[:, i] for i in range(self._lag)]

        expected_shape = data[0].shape

        for i, u in enumerate(controlin.T):
            arr = A.dot(data[i]) + self._B.dot(u)
            if arr.shape != expected_shape:
                raise ValueError(
                    f"Invalid shape: expected {expected_shape}, got {arr.shape}"
                )
            data.append(arr)

        data = np.array(data).T
        return data

    def fit(self, X, I, B=None):
        """
        Compute the Dynamic Modes Decomposition with control given the original
        snapshots and the control input data. The matrix `B` that controls how
        the control input influences the system evolution can be provided by
        the user; otherwise, it is computed by the algorithm.

        :param X: the input snapshots.
        :type X: numpy.ndarray or iterable
        :param I: the control input.
        :type I: numpy.ndarray or iterable
        :param numpy.ndarray B: matrix that controls the control input
            influences the system evolution.
        :type B: numpy.ndarray or iterable
        """
        self._reset()
        self._controlin = np.atleast_2d(np.asarray(I))

        self._snapshots_holder = Snapshots(X)
        n_samples = self.snapshots.shape[-1]

        if self._lag < 1:
            raise ValueError("Time lag must be positive.")

        X = self.snapshots[:, : -self._lag]
        Y = self.snapshots[:, self._lag :]

        self._set_initial_time_dictionary(
            {"t0": 0, "tend": n_samples - 1, "dt": 1}
        )

        if B is None:
            self._Atilde = DMDBUnknownOperator(**self._dmd_operator_kwargs)
            self._basis, self._B = self.operator.compute_operator(
                X, Y, self._controlin
            )
        else:
            self._Atilde = DMDBKnownOperator(**self._dmd_operator_kwargs)
            U, _, _ = self.operator.compute_operator(X, Y, B, self._controlin)

            self._basis = U
            self._B = B

        self._b = self._compute_amplitudes()

        return self



"""
Module for snapshots normalization.
"""

import logging
import warnings

import numpy as np


class Snapshots:
    """
    Utility class to preprocess snapshots shape for DMD.

    This class expects the time to be the last dimensions of the array.
    If a Python list is passed to the constructor, each element in the
    list is assumed to be a snapshot in time.

    Space dimensions are flattened (C-order) such that the
    matrix becomes 2D (time changes along the last axis).

    :param numpy.array | list(numpy.array) X: Training snapshots.
    """

    def __init__(self, X):
        (
            self._snapshots,
            self._snapshots_shape,
        ) = Snapshots._unroll_space_dimensions(X)

        if self._snapshots.shape[-1] == 1:
            raise ValueError("Received only one time snapshot.")

        Snapshots._check_condition_number(self._snapshots)

        logging.info(
            "Snapshots: %s, snapshot shape: %s",
            self._snapshots.shape,
            self._snapshots_shape,
        )

    @staticmethod
    def _unroll_space_dimensions(X):
        if hasattr(X, "ndim"):
            if X.ndim == 1:
                raise ValueError(
                    "Expected at least a 2D matrix (space x time)."
                )
            snapshots = X.reshape((-1, X.shape[-1]))
            shapes = set((X.shape[:-1],))
        else:
            shapes, arrays = zip(
                *[(xarr.shape, xarr.flatten()) for xarr in map(np.asarray, X)]
            )

            shapes = set(shapes)
            if len(shapes) != 1:
                raise ValueError(
                    f"Snapshots must have the same size, found {len(shapes)}."
                )
            if len(next(iter(shapes))) == 0:
                raise ValueError("Expected at least a 2D matrix")

            # move the time to the last axis
            snapshots = np.moveaxis(np.stack(arrays), 0, -1)

        return snapshots, shapes.pop()

    @staticmethod
    def _check_condition_number(X):
        cond_number = np.linalg.cond(X)
        if cond_number > 10e4:
            warnings.warn(
                f"Input data condition number {cond_number}. "
                """Consider preprocessing data, passing in augmented data
matrix, or regularization methods."""
            )

    @property
    def snapshots(self):
        """
        Snapshots of the system (space flattened).
        """
        return self._snapshots

    @property
    def snapshots_shape(self):
        """
        Original (i.e. non-flattened) snapshot shape (time is ignored).
        """
        return self._snapshots_shape
